      subroutine rock4(neqn,t,tend,h,y,f,atol,rtol,work,iwork,idid)
c ----------------------------------------------------------
c
c    Numerical solution of a (mildly) stiff system
c    of first order differential equations. ROCK4 is
c    based on a family of 4th order explicit Runge-
c    Kutta methods with nearly optimal stability domain
c    on the negative real axis. The numerical method is
c    based on a three-term recursion relation, and a
c    composition of two sub-methods.
c    The size (along the negative axis) of the stability
c    domains increases quadratically with the stage number.
c
c    Intended for problems of large dimensions with
c    eigenvalues of the Jacobian close to the negative
c    real axis. Typically for problems originating from
c    parabolic PDEs.
c
c
c
c
c    Author: A. Abdulle
c            Universite de Geneve, Dept. de mathematiques
c            Ch-1211 Geneve 24, Switzerland
c            e-mail: assyr.abdulle@math.unige.ch
c
c    Version of April 2002  minor bug corrected
c
c    The analysis of the ROCK4 method is described in:
c
c    [ ] A. Abdulle
c        Fourth order Chebyshev methods with
c        recurrence relation
c        To appear in SISC
c        http://www.unige.ch/math/biblio/preprint/liste.html
c
c     Input parameters
c     ----------------
c     NEQN:       Number of differential equations of the system
c                 (integer).
c
c     T:          Initial point of integration (double precision).
c
c     TEND:       End of the interval of integration,
c                 may be less than t (double precision)
c
c     H:          Initial step size guess
c                 (usually between 1d-4 and 1d-6).
c
c     Y(NEQN):    Initial value of the solution
c                 (double precision array of length neqn).
c
c     F:          Name (external) of subroutine computing the value
c                 of f(x,y). Must have the form
c
c                   subroutine f(neqn,t,y,dy)
c                   double precision y(neqn),dy(neqn)
c                   integer neqn
c                   dy(1)=...
c                   ...
c                   dy(neqn)=...
c                   return
c                   end
c
c                 Implementation:
c                 for stability issues when the problem
c                 is originating from parabolic PDEs, transforming
c                 inhomogeneous boundary conditions in homogeneous ones
c                 (by adding the appropriate function to the right-hand side)
c                 may increase the performance of the code.
c
c
c
c     ATOL(*) :   Absolute and relative error tolerances
c     RTOL(*)     can be both scalar (double precision)
c                 or vectors of length neqn (double precision).
c
c     RHO:        Name (external) of a function (double precision)
c                 giving the spectral radius of the Jacobian
c                 matrix of f at (t,y). Must have the form
c
c                   double precision function rho(neqn,t,y)
c                   double precision y(neqn),t
c                   integer neqn
c                   ...
c                   rho=...
c                   return
c                   end
c
c                 N.b. Gerschgorin's theorem can be helpful. If the
c                 Jacobian is known to be constant it should be
c                 specified by setting iwork(2)=1 (see below).
c
c                 ROCK4 can also compute this estimate. In that
c                 case, provide a dummy function rho(neqn,t,y)
c                 and set iwork(1)=0 (see below).
c
c                 If it is possible to give an estimate of
c                 the spectral radius, it should be preferred to
c                 the estimate computed internally by ROCK4.
c
c     IWORK(*):   Integer array of length 12 that gives information
c                 on how the problem is to be solved and communicates
c                 statistics about the integration process.
c
c     IWORK(1):   =0 ROCK4 attempts to compute the spectral radius
c                    internally. Define a dummy function
c
c                    double precision function rho(neqn,t,y)
c                    double precision y(neqn),t
c                    integer neqn
c                    rho=0.d0
c                    return
c                    end
c
c                 =1 RHO returns an upper bound of the spectral
c                    radius  of the Jacobian matrix of f at (t,y).
c
c     IWORK(2):   =0 The Jacobian is not constant.
c                 =1 The Jacobian is constant
c                    the function rho is called only once.
c
c     IWORK(3):   =0 Return and solution at tend.
c                 =1 The code returns after each step t_i chosen
c                    automatically between [t,tend] (solution
c                    at t_i is in y(*) ).
c                    To continue call ROCK4 again without changing
c                    any arguments.
c
c     IWORK(4):   =0 Atol and rtol are scalar.
c                 =1 Atol and rtol are array of length neqn.
c
c     WORK(*):      Workspace of length 8*neqn if iwork(1)=0,
c                   otherwise of length 7*neqn.
c                   Work(1),..,work(7*neqn) serve as
c                   working space for the solution of
c                   the ode.
c                   Work(7*neqn+1),..,work(8*neqn)
c                   serve as working space for the
c                   internal computation of the
c                   spectral radius of the Jacobian.
c
c     IDID:         Report on successfulness upon return
c                   (integer).
c
c
c     Output parameters
c     -----------------
c     T:          T-value for which the solution has been computed
c                 (after successful return t=tend).
c
c     Y(NEQN):    Numerical solution at tend.
c
c     IDID:       Reports what happened upon return
c
c     IDID        =1 Successful computation t=tend.
c                 =2 Successful computation of one step.
c                    To continue call ROCK4 again without
c                    altering any arguments.
c                 =-1 invalid input parameters.
c                 =-2 Stepsize becomes to small.
c                 =-3 The method used in ROCK4 to estimate
c                     the spectral radius did not converge.
c
c     IWORK(5)    =Number of function evaluations.
c     IWORK(6)    =Number of steps.
c     IWORK(7)    =Number of accepted steps.
c     IWORK(8)    =Number of rejected steps.
c     IWORK(9)    =Number of evaluations of f used
c                 to estimate the spectral radius
c                 (equal to zero if iwork(1)=1).
c     IWORK(10)   =Maximum number of stages used.
c     IWORK(11)   =Maximum value of the estimated
c                  bound for the spectral radius
c                  (rounded to the nearest integer).
c     IWORK(12)   =Minimum value of the estimated
c                  bound for the spectral radius
c                  (rounded to the nearest integer).
c
c
c    Caution:     The variable UROUND (the rounding unit) is set to
c    -------      1.0d-16 and may depends on the machines.
c-------------------------------------------------------------------
c *** *** *** *** *** *** *** *** *** *** *** *** *** *** ***
c         Numerical method
c *** *** *** *** *** *** *** *** *** *** *** *** *** *** ***
c      The nearly optimal stability polynomial is computed
c      as a product:  R_s(z)=P_{s-4}(z)*w(z).
c      We realize this polynomial as a composition of two
c      Runge-Kutta methods W(P), where the stability
c      polynomial of the method P is P_{s-4}(z) and the stability
c      polynomial of the method W is w(z). The first s-4 stages
c      possess a three-term recurrence formula.
c
c *** *** *** *** *** *** *** *** *** *** *** *** *** *** ***
c         Stability functions and three-term recurrence relation
c *** *** *** *** *** *** *** *** *** *** *** *** *** *** ***
c
c     The stability functions: R_j(z)=P_{j-4}(z)  (internal j<=ms-4)
c                             R_{ms}=P_{s-4}(z)*w(z) ( absolute)
c                             w(z)=1+a_1*z+a_2*z^2+a_3*z^3+a_4*z^4
c
c     P_j(z) orthogonal with respect to w(z)^2/sqrt{1-x^2}
c
c     Recurrence formula:
c
c     P_j(z)=(a_j*z-b_j)*P_{j-1}(z)-c_j*P_{j-2}(z)
c                 j=1..ms-4 b_1=-1,c_1=0
c
c     Normalization: P_(0)=1  =>b_{j}=-(1+c_{j})
c
c     Runge-Kutta formula:
c
c     g_j(z)=a_{j}*f(g_{j-1})-b_{j}*g_{j-1}-c_{j}*g_{j-2} j=1..ms-4
c
c     Data (rec. param.): rec(i)= a_1,a_2,c_2,a_3,c_3,.,a_(ms-4),
c                          c_(ms-4)  for ms=1,3,5,..
c
c     The finishing procedure:
c
c     g_{s-3}=g_{s-4} + h*a_{2,1}*f(g_{s-4}
c     g_{s-2}=g_{s-4} + h*(a_{3,1}*f(g_{s-4}+a_{3,2}*f(g_{s-3})
c     g_{s-1}=g_{s-4} + h*(a_{4,1}*f(g_{s-4}
c                     + a_{4,2}*f(g_{s-3}a_{4,3}*f(g_{s-2})
c     y1:=g_{s}=g_{s-4} h*(b_1*f(g_{s-4}+b_2*f(g_{s-3}+
c     b_3*f(g_{s-2})+b_4*f(g_{s-1})
c
c     Embedded method:
c
c     y1e:=g_{s}=g_{s-4} h*(be_1*f(g_{s-4}+be_2*f(g_{s-3}+
c     be_3*f(g_{s-2})+be_4*f(g_{s-1}+be_5*f(y_1))
c
c     Data (finish. proced.):fpa(i,j)=a_ij,fpb(i)=b_i,fpbe(i)=be_i
c                             i=1,.,4 j=1,.,i-1  for ms=1,3,5,..
c
c     Chosen degrees: s=5,6,7,8,9,10,11,12,13,14,15,16,17,18,19,20,21,22,23,24,
c     26,28,30,32,34,36,38,40,42,45,48,51,54,57,60,63,67,71,75,
c     80,85,90,96,102,109,116,124,133,142,152  ms=s-4
c
c------------------------------------------------------------------
c
c *** *** *** *** *** *** *** *** *** *** *** *** *** *** ***
c             Declarations
c *** *** *** *** *** *** *** *** *** *** *** *** *** *** ***
c
      double precision y(neqn),work(*),atol(*),rtol(*),
     & t,tend,h,uround,recf(4382),fpa(50,6),fpb(50,4),
     & fpbe(50,5)
      integer iwork(12),ms(50),neqn,i,j,n1,n2,n3,n4,n5,
     & n6,n7,ntol,idid
      logical arret
      external f
c -------- Uround: smallest number satisfying 1.d0+uround>1.d0
      data uround/1.0d-16/
c
c *** *** *** *** *** *** *** *** *** *** *** *** *** *** ***
c             Data of the stability polynomials
c *** *** *** *** *** *** *** *** *** *** *** *** *** *** ***
c -------- Degrees ----------------------------------------------------
       data ms/1,2,3,4,5,6,7,8,9,10,11,12,13,14,15,16,17,18,19,20,
     & 22,24,26,28,30,32,34,36,38,41,44,47,50,53,56,59,63,67,71,
     & 76,81,86,92,98,105,112,120,129,138,148/
c -------- Parameters of the finishing procedure.----------------------
      data ((fpa(i,j),j=1,6),i=1,9)/
     &-.149352078672699D+00,0.629768962985252D+00,-.355201061573650D+00,
     &0.146745996307541D-01,-.558517281602565D-01,0.590312931352706D+00,
     &-.179313593132677D+00,0.522354978881996D+00,-.268020263243268D+00,
     &0.410644459848542D-01,-.673692676033641D-01,0.534974252895965D+00,
     &-.182428104178289D+00,0.488581974232839D+00,-.243999027372673D+00,
     &0.403844469216897D-01,-.621100542563179D-01,0.510891501054959D+00,
     &-.177747282156689D+00,0.475598890858970D+00,-.236544145857569D+00,
     &0.357873367680739D-01,-.561802405454197D-01,0.498502393780149D+00,
     &-.171097214294784D+00,0.471021380996139D+00,-.235432301175529D+00,
     &0.305165148305403D-01,-.508758913107147D-01,0.491537536121394D+00,
     &-.164446145110292D+00,0.470315341552133D+00,-.237054736024656D+00,
     &0.253636332272387D-01,-.462365759807733D-01,0.487394153808489D+00,
     &-.158427655171257D+00,0.471468878620055D+00,-.239853802450812D+00,
     &0.205746749952032D-01,-.421744356707249D-01,0.484829913014008D+00,
     &-.153293058075572D+00,0.473282079177035D+00,-.242910850765406D+00,
     &0.163722321103915D-01,-.387108095861697D-01,0.483081034299037D+00,
     &-.148972804305139D+00,0.475296959544537D+00,-.245899260153289D+00,
     &0.127203010191100D-01,-.357508839759137D-01,0.481825981739298D+00/
      data ((fpa(i,j),j=1,6),i=10,18)/
     &-.145179937599982D+00,0.477661231491125D+00,-.248988404183759D+00,
     &0.930730703798092D-02,-.330359105371527D-01,0.481074258113903D+00,
     &-.141803779162182D+00,0.480294472251598D+00,-.252149638985389D+00,
     &0.605980548716020D-02,-.304878190339923D-01,0.480717680079251D+00,
     &-.139121943562170D+00,0.482307569354585D+00,-.254656989245024D+00,
     &0.349723479360248D-02,-.284636204606306D-01,0.480267545886151D+00,
     &-.136626653539895D+00,0.484641896383759D+00,-.257332057712838D+00,
     &0.915869293860006D-03,-.264509129364182D-01,0.480154720984400D+00,
     &-.134486447576714D+00,0.486757723660692D+00,-.259733624852103D+00,
     &-.135988997170157D-02,-.246765061257936D-01,0.480084593714673D+00,
     &-.132780674018083D+00,0.488321952587049D+00,-.261588769418801D+00,
     &-.313701049350723D-02,-.232786280695929D-01,0.479882004899598D+00,
     &-.131117609799413D+00,0.490193639754613D+00,-.263646432339284D+00,
     &-.502134729674226D-02,-.218133033112692D-01,0.479929065438669D+00,
     &-.129737943206222D+00,0.491700319201409D+00,-.265337954300747D+00,
     &-.657602712355092D-02,-.205977998576123D-01,0.479898556782486D+00,
     &-.128546948141681D+00,0.493011323225514D+00,-.266816177922376D+00,
     &-.793004771706004D-02,-.195369613099883D-01,0.479857299633325D+00/
      data ((fpa(i,j),j=1,6),i=19,27)/
     &-.127460377104448D+00,0.494293149884316D+00,-.268231271325724D+00,
     &-.920717061172173D-02,-.185386009080156D-01,0.479869528636923D+00,
     &-.126496410155342D+00,0.495458691659965D+00,-.269513232370108D+00,
     &-.103575442717928D-01,-.176387115527810D-01,0.479887174404288D+00,
     &-.124868416118615D+00,0.497490597051425D+00,-.271737391165211D+00,
     &-.123394260182449D-01,-.160868868427959D-01,0.479932724633469D+00,
     &-.123564405510046D+00,0.499164732520327D+00,-.273565616640958D+00,
     &-.139585436258901D-01,-.148169722656038D-01,0.479973747650230D+00,
     &-.122505008329286D+00,0.500555383947992D+00,-.275081797679137D+00,
     &-.152951780850662D-01,-.137670777523960D-01,0.480009428375171D+00,
     &-.121640329449530D+00,0.501700402349954D+00,-.276333007672789D+00,
     &-.163961730000813D-01,-.129007107727469D-01,0.480031673127157D+00,
     &-.120917917691482D+00,0.502675119491259D+00,-.277395466055127D+00,
     &-.173277807640047D-01,-.121670308408281D-01,0.480054118477096D+00,
     &-.120710248396632D+00,0.502361065588637D+00,-.277299180400763D+00,
     &-.173459597107178D-01,-.121268146405534D-01,0.479596544727020D+00,
     &-.120151946726542D+00,0.503187463625006D+00,-.278173489088946D+00,
     &-.180998664145073D-01,-.115355117518033D-01,0.479663327238431D+00/
      data ((fpa(i,j),j=1,6),i=28,36)/
     &-.119650804512964D+00,0.503968844736113D+00,-.278987558251327D+00,
     &-.187954561730573D-01,-.109907774512063D-01,0.479748806593835D+00,
     &-.119231690274109D+00,0.504610327555198D+00,-.279662019452110D+00,
     &-.193736976241804D-01,-.105368584242475D-01,0.479807172254604D+00,
     &-.118685223311611D+00,0.505479309944796D+00,-.280566027535941D+00,
     &-.201436607277488D-01,-.993291024114479D-02,0.479903135786603D+00,
     &-.118231796286135D+00,0.506213890671776D+00,-.281327499307895D+00,
     &-.207902381330527D-01,-.942553811892237D-02,0.479988558979737D+00,
     &-.117860168810132D+00,0.506813986559461D+00,-.281951733666353D+00,
     &-.213206692268858D-01,-.900871675159034D-02,0.480053891764692D+00,
     &-.117541632407920D+00,0.507340547092223D+00,-.282496164808253D+00,
     &-.217814864430309D-01,-.864670120809640D-02,0.480116952219068D+00,
     &-.117280791290490D+00,0.507762024730383D+00,-.282936101835780D+00,
     &-.221553810871189D-01,-.835235476491614D-02,0.480159581641243D+00,
     &-.117020809344219D+00,0.508234853606851D+00,-.283411245544839D+00,
     &-.225511383509412D-01,-.804255449961079D-02,0.480240908974576D+00,
     &-.116840110106662D+00,0.508513931927332D+00,-.283708061860511D+00,
     &-.228054436800931D-01,-.784157438541911D-02,0.480258758199155D+00/
      data ((fpa(i,j),j=1,6),i=37,45)/
     &-.116999141776097D+00,0.507738965363869D+00,-.283085550597818D+00,
     &-.223557928014705D-01,-.817646181729438D-02,0.479839084150866D+00,
     &-.116787474223622D+00,0.508124510062097D+00,-.283473179991593D+00,
     &-.226786360708563D-01,-.792370493215401D-02,0.479905001144018D+00,
     &-.116543053242319D+00,0.508649023059358D+00,-.283975543942050D+00,
     &-.230858609767789D-01,-.760742817737876D-02,0.480040247388773D+00,
     &-.116369877848172D+00,0.508950670881779D+00,-.284284071372858D+00,
     &-.233448484372084D-01,-.740387836067644D-02,0.480081925815726D+00,
     &-.115981831952477D+00,0.509942937789055D+00,-.285192892337194D+00,
     &-.240615204888854D-01,-.685134671736066D-02,0.480412958109969D+00,
     &-.115934409239879D+00,0.509932625794726D+00,-.285214065725890D+00,
     &-.240929036649298D-01,-.682325905888604D-02,0.480353282861487D+00,
     &-.114543930853965D+00,0.514055666969259D+00,-.288877128663596D+00,
     &-.269187182378037D-01,-.465261555222027D-02,0.481928410401348D+00,
     &-.114338190892959D+00,0.514568129329710D+00,-.289354467784058D+00,
     &-.272970253684862D-01,-.435838305221696D-02,0.482082731512007D+00,
     &-.113900806706479D+00,0.515815247278763D+00,-.290479621681517D+00,
     &-.281706242269155D-01,-.368286974902504D-02,0.482524745170433D+00/
      data ((fpa(i,j),j=1,6),i=46,50)/
     &-.114241230745559D+00,0.514642605930859D+00,-.289464730587169D+00,
     &-.274037778296787D-01,-.427016797791793D-02,0.482029696494976D+00,
     &-.113830853818569D+00,0.515834254672657D+00,-.290535565769687D+00,
     &-.282330735764243D-01,-.362939602828681D-02,0.482459847410651D+00,
     &-.113105039346215D+00,0.518047629039909D+00,-.292508542364670D+00,
     &-.297505373346793D-01,-.245703714012197D-02,0.483285747825278D+00,
     &-.113373695071906D+00,0.517118156325443D+00,-.291701320745508D+00,
     &-.291406743553601D-01,-.292560575875699D-02,0.482899951273987D+00,
     &-.112938370080633D+00,0.518440212713205D+00,-.292882869228272D+00,
     &-.300502325911247D-01,-.222188377341845D-02,0.483386803334408D+00/
c
      data ((fpb(i,j),j=1,4),i=1,12)/
     &0.934502625489809D+00,-.426556402801135D+00,-.428612609028723D+00,
     &0.744370090574855D+00,0.727428506710284D+00,-.281882500610754D+00,
     &-.366883181811475D+00,0.684341322628128D+00,0.671969568061243D+00,
     &-.247327027296030D+00,-.348410422675722D+00,0.657516722491004D+00,
     &0.655616700297981D+00,-.238808679740035D+00,-.342918097601918D+00,
     &0.643274312048176D+00,0.653766630051757D+00,-.239717659625342D+00,
     &-.342124625373560D+00,0.634842894408973D+00,0.657958401550080D+00,
     &-.244504571310874D+00,-.343125007747822D+00,0.629452994091047D+00,
     &0.664644828327504D+00,-.250836131475691D+00,-.344767760715637D+00,
     &0.625804292371554D+00,0.671778926897883D+00,-.257341094888494D+00,
     &-.346437936729090D+00,0.623113789954589D+00,0.678662206121957D+00,
     &-.263538563990534D+00,-.347971227180309D+00,0.621040683222628D+00,
     &0.685711960310753D+00,-.269724027269246D+00,-.349535907736682D+00,
     &0.619566456617273D+00,0.692884607363355D+00,-.275886628173354D+00,
     &-.351131980732774D+00,0.618568501341401D+00,0.698546656891454D+00,
     &-.280881006247969D+00,-.352285102713857D+00,0.617571192399056D+00/
      data ((fpb(i,j),j=1,4),i=13,24)/
     &0.704573921792281D+00,-.286045457291729D+00,-.353563979295968D+00,
     &0.616965030808178D+00,0.709971428905986D+00,-.290682008221983D+00,
     &-.354679184547433D+00,0.616462060289197D+00,0.714125178664410D+00,
     &-.294349691347616D+00,-.355464510591820D+00,0.615888572779772D+00,
     &0.718736274563484D+00,-.298288036358834D+00,-.356401512165052D+00,
     &0.615594896206387D+00,0.722517387767685D+00,-.301568462204942D+00,
     &-.357129478686764D+00,0.615267647826005D+00,0.725816343813274D+00,
     &-.304448113195281D+00,-.357746756688570D+00,0.614963961979991D+00,
     &0.728974076820863D+00,-.307182320589777D+00,-.358345483491417D+00,
     &0.614739362936109D+00,0.731832397913825D+00,-.309659464712225D+00,
     &-.358881456195461D+00,0.614544900863432D+00,0.736786729484850D+00,
     &-.313957288328657D+00,-.359797924432667D+00,0.614228100935116D+00,
     &0.740854099130286D+00,-.317495426161459D+00,-.360535121685962D+00,
     &0.613973796355241D+00,0.744223896195403D+00,-.320433632193417D+00,
     &-.361135560093482D+00,0.613766054898060D+00,0.747001832700433D+00,
     &-.322865419691933D+00,-.361620350229670D+00,0.613586121252665D+00/
      data ((fpb(i,j),j=1,4),i=25,36)/
     &0.749359431920161D+00,-.324930741701897D+00,-.362028107948934D+00,
     &0.613438378039065D+00,0.749100998320598D+00,-.324985993239383D+00,
     &-.361796607662197D+00,0.612867258144605D+00,0.751043873200700D+00,
     &-.326661586308774D+00,-.362147816931696D+00,0.612807453647952D+00,
     &0.752854624396087D+00,-.328210433831008D+00,-.362482431451812D+00,
     &0.612782100341090D+00,0.754353169889782D+00,-.329501139703409D+00,
     &-.362752509396770D+00,0.612745403519662D+00,0.756363017447499D+00,
     &-.331222786188901D+00,-.363119850550614D+00,0.612719466518580D+00,
     &0.758055963580132D+00,-.332671506824595D+00,-.363429198770158D+00,
     &0.612703916106263D+00,0.759443004727975D+00,-.333862330033422D+00,
     &-.363679324780792D+00,0.612685408765563D+00,0.760653160327002D+00,
     &-.334898139570366D+00,-.363899175365336D+00,0.612677301460611D+00,
     &0.761629956010211D+00,-.335739950284413D+00,-.364072376423153D+00,
     &0.612660139381166D+00,0.762688789887627D+00,-.336630474589357D+00,
     &-.364274785126092D+00,0.612687294013859D+00,0.763346400554849D+00,
     &-.337204680067583D+00,-.364385896133756D+00,0.612661785846953D+00/
      data ((fpb(i,j),j=1,4),i=37,48)/
     &0.761923304846226D+00,-.336206156720779D+00,-.363978603432088D+00,
     &0.612221699604794D+00,0.762786806785331D+00,-.336933071907623D+00,
     &-.364142956905108D+00,0.612243212238912D+00,0.763911644345184D+00,
     &-.337848624997397D+00,-.364378249037553D+00,0.612335667041690D+00,
     &0.764597562248984D+00,-.338433363367646D+00,-.364503441229143D+00,
     &0.612339040874567D+00,0.766642781170500D+00,-.340044276021276D+00,
     &-.364967251951014D+00,0.612618883157375D+00,0.766682555493412D+00,
     &-.340118220148353D+00,-.364946973037536D+00,0.612538317898983D+00,
     &0.774955438746513D+00,-.346489297090328D+00,-.366916081411072D+00,
     &0.613985554671875D+00,0.776027351639685D+00,-.347348043660042D+00,
     &-.367147411126057D+00,0.614109087783370D+00,0.778563137974742D+00,
     &-.349330512988104D+00,-.367727843519890D+00,0.614502095324990D+00,
     &0.776264664166250D+00,-.347594735539935D+00,-.367159451116240D+00,
     &0.614023148520995D+00,0.778679038214519D+00,-.349476481579656D+00,
     &-.367715953710640D+00,0.614409463784688D+00,0.783131247138323D+00,
     &-.352927933064792D+00,-.368752741091226D+00,0.615166687043411D+00/
      data ((fpb(i,j),j=1,4),i=49,50)/
     &0.781304104495501D+00,-.351541585232233D+00,-.368306700790280D+00,
     &0.614794688266817D+00,0.783969323650751D+00,-.353613571196170D+00,
     &-.368922535023076D+00,0.615238813023293D+00/
c
      data ((fpbe(i,j),j=1,5),i=1,9)/
     &0.113509972110540D+01,-.584333360989720D+00,-.319172911177732D+00,
     &0.482853558185876D+00,0.109256697110981D+00,0.100479199674788D+01,
     &-.436644602461537D+00,-.423443966877378D+00,0.538770645909717D+00,
     &0.795300735974956D-01,0.984907410167693D+00,-.403297625005343D+00,
     &-.485090625106059D+00,0.576309835350610D+00,0.609198451735942D-01,
     &0.980171201303714D+00,-.394437889689979D+00,-.517381582925487D+00,
     &0.601142288908999D+00,0.476702174069568D-01,0.974834695776200D+00,
     &-.392308952799017D+00,-.530263724070033D+00,0.616510618632796D+00,
     &0.379946019218839D-01,0.966668353329992D+00,-.391741144639394D+00,
     &-.531455776704021D+00,0.625494551194751D+00,0.308158334011040D-01,
     &0.956319747720222D+00,-.391138376736994D+00,-.526142643689543D+00,
     &0.630404554467362D+00,0.254019467466823D-01,0.943587787064690D+00,
     &-.389377300510789D+00,-.516876350192269D+00,0.632531861517858D+00,
     &0.212476873553974D-01,0.929868441621807D+00,-.386750527439590D+00,
     &-.505935645060380D+00,0.633001592986906D+00,0.180092360649997D-01/
      data ((fpbe(i,j),j=1,5),i=10,18)/
     &0.917878947092234D+00,-.384686528119553D+00,-.495511556683115D+00,
     &0.632891756120139D+00,0.154458635123923D-01,0.907921031017813D+00,
     &-.383354597608536D+00,-.486086276370317D+00,0.632568615019003D+00,
     &0.133857277406643D-01,0.896388476009400D+00,-.380504535913835D+00,
     &-.476131894672218D+00,0.631491892380436D+00,0.117078025249012D-01,
     &0.887685870010584D+00,-.378944197433540D+00,-.467828653690999D+00,
     &0.630692391160778D+00,0.103241059659401D-01,0.879470814062224D+00,
     &-.377245094706344D+00,-.460102121012690D+00,0.629778007382364D+00,
     &0.917069070021292D-02,0.870530751362587D+00,-.374648589005799D+00,
     &-.452452352164561D+00,0.628569291738020D+00,0.820044757449899D-02,
     &0.864018892596866D+00,-.373280954775127D+00,-.446202255714937D+00,
     &0.627730565946685D+00,0.737537419249938D-02,0.857357924295629D+00,
     &-.371485680656218D+00,-.440234425775788D+00,0.626780133051391D+00,
     &0.666914378696866D-02,0.851096961215784D+00,-.369663988413454D+00,
     &-.434754090977242D+00,0.625846657124080D+00,0.605989696024640D-02/
      data ((fpbe(i,j),j=1,5),i=19,27)/
     &0.845691163494378D+00,-.368162878390094D+00,-.429905486466339D+00,
     &0.625032425385260D+00,0.553041165257352D-02,0.840754783321373D+00,
     &-.366754374490097D+00,-.425504080282415D+00,0.624272520855683D+00,
     &0.506752846502820D-02,0.832129732678031D+00,-.364215201885908D+00,
     &-.417868687835329D+00,0.622912972095680D+00,0.430080260616755D-02,
     &0.824841358445432D+00,-.361969316759517D+00,-.411506554101034D+00,
     &0.621735648069777D+00,0.369621198344728D-02,0.818660702021942D+00,
     &-.359997706140366D+00,-.406173910081373D+00,0.620720634622955D+00,
     &0.321103838340649D-02,0.813343156361523D+00,-.358234931076463D+00,
     &-.401657773073081D+00,0.619835953653306D+00,0.281577816620828D-02,
     &0.808812551334635D+00,-.356709737514645D+00,-.397829666549206D+00,
     &0.619076368359449D+00,0.248944467816156D-02,0.802058241484851D+00,
     &-.353305127983149D+00,-.393665858753007D+00,0.617880189866334D+00,
     &0.221821094859412D-02,0.799023019885202D+00,-.352371068280062D+00,
     &-.390970958156562D+00,0.617372975124671D+00,0.198795503493263D-02/
      data ((fpbe(i,j),j=1,5),i=28,36)/
     &0.796533707522770D+00,-.351659214274525D+00,-.388680577600314D+00,
     &0.616958145987822D+00,0.179179781860265D-02,0.794254415802991D+00,
     &-.350955011234794D+00,-.386652430547127D+00,0.616574541234043D+00,
     &0.162340905415174D-02,0.791451974069317D+00,-.350127723737597D+00,
     &-.384099751306937D+00,0.616103218945433D+00,0.141212925634866D-02,
     &0.789140296939903D+00,-.349447391105024D+00,-.381986708309451D+00,
     &0.615713343205178D+00,0.123963336103694D-02,0.787155036742462D+00,
     &-.348839107260454D+00,-.380202863720354D+00,0.615376708082511D+00,
     &0.109698483515876D-02,0.785509443687073D+00,-.348347728366112D+00,
     &-.378703894403193D+00,0.615097681507449D+00,0.977644426693831D-03,
     &0.784038223206202D+00,-.347876842227869D+00,-.377407083344177D+00,
     &0.614846655967241D+00,0.876815082414696D-03,0.783010378301710D+00,
     &-.347648245486129D+00,-.376356876713355D+00,0.614674780335034D+00,
     &0.790787748777402D-03,0.781833266343420D+00,-.347234482878725D+00,
     &-.375370253527294D+00,0.614472179428255D+00,0.716900834807566D-03/
      data ((fpbe(i,j),j=1,5),i=37,45)/
     &0.778157258445688D+00,-.345006104027160D+00,-.373629716648469D+00,
     &0.613804905745683D+00,0.633900782409020D-03,0.777300034873279D+00,
     &-.344806720424309D+00,-.372764739426396D+00,0.613661241678936D+00,
     &0.564173510000384D-03,0.776985336763925D+00,-.344948310351192D+00,
     &-.372138199888685D+00,0.613616309802328D+00,0.505301025547597D-03,
     &0.776123576908821D+00,-.344696744011525D+00,-.371340582024658D+00,
     &0.613469644712914D+00,0.443902941209962D-03,0.776949282583553D+00,
     &-.345654199916775D+00,-.371072531043118D+00,0.613634699306968D+00,
     &0.392885424956897D-03,0.775876610212432D+00,-.345123309087147D+00,
     &-.370392356339381D+00,0.613444359409516D+00,0.350376011087039D-03,
     &0.783347589653809D+00,-.351086251339795D+00,-.371860974551612D+00,
     &0.614828131953302D+00,0.307119201283797D-03,0.783496656673086D+00,
     &-.351443051706240D+00,-.371545102252444D+00,0.614860564019234D+00,
     &0.271917903319380D-03,0.785178130625954D+00,-.352964147655529D+00,
     &-.371615967722576D+00,0.615170969674100D+00,0.237891869788885D-03/
      data ((fpbe(i,j),j=1,5),i=46,50)/
     &0.782045229930799D+00,-.350764936286095D+00,-.370561648626977D+00,
     &0.614604810918108D+00,0.210170095234033D-03,0.783792445229447D+00,
     &-.352285904984906D+00,-.370720709266521D+00,0.614926468923107D+00,
     &0.183766807783551D-03,0.787664243592327D+00,-.355426631067992D+00,
     &-.371408773264980D+00,0.615628921926696D+00,0.159498839664580D-03,
     &0.785248632105467D+00,-.353713152595833D+00,-.370620439974436D+00,
     &0.615195478568662D+00,0.139988635943928D-03,0.787452369208138D+00,
     &-.355534844520464D+00,-.370962019539811D+00,0.615594462312992D+00,
     &0.122062993942758D-03/
c ------- Recurrence parameters.------------------------------------
      data (recf(i),i=1,56)/.1762962957651941D+00,.1067131485543800D+00,
     &.1296319693918818D+00,.6097984609869632D-02,.7296375286357569D-01,
     &.8749824867149327D-01,.6126949805354446D-02,.1026802330400501D+00,
     &.3026744779900850D-01,.5345766591735225D-01,.6421114488497465D-01,
     &.6297251335072349D-02,.7361067001944596D-01,.2954244012876854D-01,
     &.8450213007949955D-01,.6370198138618569D-01,.4102541073974864D-01,
     &.4936734181284644D-01,.6437548414781547D-02,.5655964819916140D-01,
     &.3008674068005620D-01,.6301337991392047D-01,.6092514503458917D-01,
     &.7122116739390528D-01,.1014508394981259D+00,.3256098772501858D-01,
     &.3923994550846150D-01,.6548791582739922D-02,.4499511886194148D-01,
     &.3061034821432499D-01,.5000269311882022D-01,.6157339578289546D-01,
     &.5465516336291797D-01,.9570538881974760D-01,.6104333020232396D-01,
     &.1406290220864037D+00,.2651417777294218D-01,.3199100352247176D-01,
     &.6637416338030012D-02,.3671473048471728D-01,.3104646653985547D-01,
     &.4080808506561111D-01,.6240780397840483D-01,.4441697436286914D-01,
     &.9608256539833336D-01,.4790790953491602D-01,.1314940925346221D+00,
     &.5299225195731577D-01,.1795490141555294D+00,.2203688946407306D-01,
     &.2661454798174795D-01,.6708509073973803D-02,.3056730958573840D-01,
     &.3140468274131318D-01,.3399153986085689D-01,.6314347647488826D-01/
       data (recf(i),i= 57,111)/                  .3698314358022203D-01,
     &.9706950916624359D-01,.3966494820191785D-01,.1312524251550016D+00,
     &.4237605842481057D-01,.1670258610104497D+00,.4651311259978263D-01,
     &.2174924541420837D+00,.1862250741526137D-01,.2250861576538788D-01,
     &.6766202022280821D-02,.2586805882924489D-01,.3170042346032810D-01,
     &.2877953008086195D-01,.6376714023863055D-01,.3131867762450776D-01,
     &.9801783524123061D-01,.3355864056201850D-01,.1322218594416999D+00,
     &.3560688231225300D-01,.1658724905539629D+00,.3777371673577270D-01,
     &.2016177039517425D+00,.4119698781191684D-01,.2539572842035754D+00,
     &.1595000101933615D-01,.1929098080329146D-01,.6813771091892661D-02,
     &.2218226060548395D-01,.3194763050768121D-01,.2468951849041972D-01,
     &.6429708643843554D-01,.2687566748685502D-01,.9885398134767074D-01,
     &.2879687047315825D-01,.1332911334324939D+00,.3051024265050759D-01,
     &.1666672556601528D+00,.3211180339134084D-01,.1993212628717513D+00,
     &.3387840798530223D-01,.2348117800966209D+00,.3672204698847085D-01,
     &.2881723275817451D+00,.1381614369504127D-01,.1671925338932986D-01,
     &.6853396714797705D-02,.1923405366588467D-01,.3215586472031486D-01,
     &.2141625996563439D-01,.6474907548155117D-01,.2331937590115891D-01,
     &.9958008772514044D-01,.2499024766114432D-01,.1342748583683893D+00/
       data (recf(i),i=112,166)/                  .2647050763871464D-01,
     &.1677665086248348D+00,.2780637354179997D-01,.1998070962536277D+00,
     &.2908398701850636D-01,.2312827828857880D+00,.3054428937166170D-01,
     &.2663434084812225D+00,.3290759457513454D-01,.3198421010321994D+00,
     &.1209208337462717D-01,.1463956211973881D-01,.6886319232429428D-02,
     &.1684821591214353D-01,.3233043835607318D-01,.1876598514920064D-01,
     &.6513176017688395D-01,.2043905166732186D-01,.1002021458157551D+00,
     &.2190770330626050D-01,.1351369448238108D+00,.2320633429617816D-01,
     &.1688220571485906D+00,.2436655982458731D-01,.2008416611727759D+00,
     &.2542750723914090D-01,.2313613787046680D+00,.2646792773501215D-01,
     &.2616749611549993D+00,.2769967463189309D-01,.2964076499426887D+00,
     &.2970440766084731D-01,.3502117812643117D+00,.1067034715302655D-01,
     &.1292329710584654D-01,.6914320860943588D-02,.1487812037525849D-01,
     &.3248006428304434D-01,.1657652594868420D-01,.6546254336130322D-01,
     &.1805880263820550D-01,.1007449363462607D+00,.1936010074490956D-01,
     &.1358990057232774D+00,.2051007922030995D-01,.1697855056821405D+00,
     &.2153416689781703D-01,.2019295372474008D+00,.2245709430196045D-01,
     &.2322718220467692D+00,.2331275056513504D-01,.2612402870437151D+00,
     &.2417201314765524D-01,.2904095866946978D+00,.2521646587374857D-01/
       data (recf(i),i=167,221)/                  .3246308095620147D+00,
     &.2690764435467064D-01,.3778240223542354D+00,.9487431939295813D-02,
     &.1149440138138981D-01,.6938125808908258D-02,.1323700105262745D-01,
     &.3260808020167218D-01,.1475188210924817D-01,.6574755356940067D-01,
     &.1607452560476730D-01,.1012162270929412D+00,.1723591531771676D-01,
     &.1365667423922086D+00,.1826211676183441D-01,.1706429091716792D+00,
     &.1917496536345655D-01,.2029443389633356D+00,.1999362278384748D-01,
     &.2333350798953319D+00,.2073811037471499D-01,.2619641661449880D+00,
     &.2143841440760416D-01,.2894413776906615D+00,.2215820736295387D-01,
     &.3175489053704050D+00,.2305270619849904D-01,.3512439576972030D+00,
     &.2449014264738941D-01,.4036179314586852D+00,.8494564197695063D-02,
     &.1029439196351078D-01,.6958406921839354D-02,.1185809158880236D-01,
     &.3271773184277712D-01,.1321815865055247D-01,.6599313162331371D-01,
     &.1440610601409755D-01,.1016249390468336D+00,.1544947770682643D-01,
     &.1371499803680846D+00,.1637143556780157D-01,.1713990070155131D+00,
     &.1719123774995229D-01,.2038574724036485D+00,.1792514377153798D-01,
     &.2343593316805466D+00,.1858798779956469D-01,.2629560717194659D+00,
     &.1919659343371288D-01,.2899296316580346D+00,.1977810373828767D-01,
     &.3160315643000932D+00,.2039024270914711D-01,.3432192271881122D+00/
       data (recf(i),i=222,276)/                  .2116722465269759D-01,
     &.3765294846983478D+00,.2240833445471075D-01,.4283514372117345D+00,
     &.7648100953017721D-02,.9270847111494889D-02,.6976054938079860D-02,
     &.1068147039692487D-01,.3281359708712486D-01,.1190898600388846D-01,
     &.6620896361366048D-01,.1298157114442065D-01,.1019862089328133D+00,
     &.1392387354640439D-01,.1376687444923527D+00,.1475662479151471D-01,
     &.1720764656800079D+00,.1549702171989481D-01,.2046849615229637D+00,
     &.1615937354505817D-01,.2353135758405889D+00,.1675598207974404D-01,
     &.2639750061864973D+00,.1729863164965770D-01,.2908276346191939D+00,
     &.1780176282268598D-01,.3162404475076796D+00,.1828983612841132D-01,
     &.3410641528662292D+00,.1881415834433531D-01,.3673520724841440D+00,
     &.1948852933531380D-01,.4000529139809629D+00,.2055287002810484D-01,
     &.4505930922691549D+00,.6923848173158734D-02,.8394702950361005D-02,
     &.6991325631797234D-02,.9673911522111265D-02,.3289687833321025D-01,
     &.1078755757084805D-01,.6639729788616958D-01,.1176100014874140D-01,
     &.1023029830546846D+00,.1261643557454879D-01,.1381259991601194D+00,
     &.1337254080889024D-01,.1726770746288135D+00,.1404479965650065D-01,
     &.2054241214472704D+00,.1464602837670213D-01,.2361779701904081D+00,
     &.1518698551280871D-01,.2649344026216049D+00,.1567717541900471D-01/
       data (recf(i),i=277,331)/                  .2918033340368942D+00,
     &.1612624123496420D-01,.3170079069341164D+00,.1654688879900881D-01,
     &.3409811920927984D+00,.1696136666824286D-01,.3646547307991162D+00,
     &.1741546145080310D-01,.3901537483439419D+00,.1800636597179543D-01,
     &.4223218228522266D+00,.1892881341268367D-01,.4717326233521751D+00,
     &.6298588439596636D-02,.7638030626687965D-02,.7004655851265844D-02,
     &.8803453943265294D-02,.3296982845467406D-01,.9818444950365933D-02,
     &.6656291057106479D-01,.1070595626475817D-01,.1025827278785117D+00,
     &.1148608766389635D-01,.1385316612194638D+00,.1217575810897931D-01,
     &.1732125747484649D+00,.1278899412440368D-01,.2060869881923743D+00,
     &.1333738432024735D-01,.2369597744170469D+00,.1383056388512692D-01,
     &.2658180187652428D+00,.1427674437223498D-01,.2927507447256686D+00,
     &.1468343958863289D-01,.3179214676099940D+00,.1505875998976432D-01,
     &.3415951393087426D+00,.1541407984395701D-01,.3642544652079210D+00,
     &.1576969712450036D-01,.3868951368176343D+00,.1616644053575092D-01,
     &.4116861214445555D+00,.1668748198434585D-01,.4433331570075596D+00,
     &.1749188724059570D-01,.4916033800451471D+00,.5754360678751824D-02,
     &.6979206263458857D-02,.7016389205940601D-02,.8045335113843407D-02,
     &.3303423464885706D-01,.8974185166510452D-02,.6670962735009836D-01/
       data (recf(i),i=332,386)/                  .9786634336394069D-02,
     &.1028314890330009D+00,.1050097258572427D-01,.1388938643805943D+00,
     &.1113260003649783D-01,.1736928024956524D+00,.1169428623074936D-01,
     &.2066843246414360D+00,.1219657523409050D-01,.2376686083847098D+00,
     &.1264820000840307D-01,.2666274169278879D+00,.1305648757877945D-01,
     &.2936397261253015D+00,.1342781257742339D-01,.3188429723296673D+00,
     &.1376825401666917D-01,.3424305607919373D+00,.1408478812041018D-01,
     &.3646959159353649D+00,.1438768989133101D-01,.3861600014637602D+00,
     &.1469542216481714D-01,.4078642344326945D+00,.1504419066763009D-01,
     &.4319869811400004D+00,.1550494890247615D-01,.4630491704469774D+00,
     &.1620775053166640D-01,.5100139240423402D+00,.5278040811520425D-02,
     &.6402421198476983D-02,.7026749310824704D-02,.7381444032562888D-02,
     &.3309125386700502D-01,.8234685519924101D-02,.6683990648107069D-01,
     &.8981222939541469D-02,.1030531116670419D+00,.9637773923336657D-02,
     &.1392177117208699D+00,.1021841796444464D-01,.1741238287384007D+00,
     &.1073483180895483D-01,.2072226942408981D+00,.1119665624742548D-01,
     &.2383105498941010D+00,.1161186412382382D-01,.2673653866792523D+00,
     &.1198709958412187D-01,.2944607083863120D+00,.1232800940556657D-01,
     &.3197217653056301D+00,.1263963481803662D-01,.3433101656487968D+00/
       data (recf(i),i=387,441)/                  .1292700918588335D-01,
     &.3654364911202174D+00,.1319625157476125D-01,.3864151426812009D+00,
     &.1345671217793918D-01,.4067976665874104D+00,.1372517143690529D-01,
     &.4276578990766819D+00,.1403368592830970D-01,.4511616717492440D+00,
     &.1444284729184994D-01,.4816264620225250D+00,.1505946839641039D-01,
     &.5272614695163532D+00,.4487666665675140D-02,.5444990141910759D-02,
     &.7044130283096086D-02,.6279049188091140D-02,.3318723192008640D-01,
     &.7006368632847780D-02,.6706002621750721D-01,.7643068862200888D-02,
     &.1034291291409337D+00,.8203282681027361D-02,.1397696656221023D+00,
     &.8698916419500412D-02,.1748620369064401D+00,.9139847522019589D-02,
     &.2081495587015637D+00,.9534232617493163D-02,.2394220617811718D+00,
     &.9888812556097355D-02,.2686520584410767D+00,.1020918148777924D-01,
     &.2959071488424056D+00,.1050001965367601D-01,.3213027967445414D+00,
     &.1076530680194359D-01,.3449782554868110D+00,.1100855051222504D-01,
     &.3670882760306343D+00,.1123309242673613D-01,.3878099745344409D+00,
     &.1144260835071195D-01,.4073711108103572D+00,.1164201330179298D-01,
     &.4261157327362403D+00,.1183914191832676D-01,.4446387149497601D+00,
     &.1204780719832178D-01,.4640454326093743D+00,.1229306401377462D-01,
     &.4864269389633110D+00,.1261927726417314D-01,.5156669165861498D+00/
       data (recf(i),i=442,496)/                  .1309943489239065D-01,
     &.5586255185473764D+00,.3862975224403902D-02,.4687945053900655D-02,
     &.7058035858061096D-02,.5407041834964283D-02,.3326430524210259D-01,
     &.6034411186094848D-02,.6723753885455682D-01,.6583863333832890D-02,
     &.1037337930420006D+00,.7067509494777436D-02,.1402191813726617D+00,
     &.7495554804930935D-02,.1754665679972444D+00,.7876466939414612D-02,
     &.2089130635289459D+00,.8217239718909702D-02,.2403434536707747D+00,
     &.8523653399761676D-02,.2697261364777181D+00,.8800501778860810D-02,
     &.2971251237465357D+00,.9051782217229110D-02,.3226519291781493D+00,
     &.9280855010852164D-02,.3464394048134897D+00,.9490584189896422D-02,
     &.3686285665457733D+00,.9683478271871005D-02,.3893644369335397D+00,
     &.9861860805271805D-02,.4088004145802082D+00,.1002812139311295D-01,
     &.4271141176574624D+00,.1018513471380169D-01,.4445422442202590D+00,
     &.1033699621003472D-01,.4614490459910935D+00,.1049031549106710D-01,
     &.4784539901229190D+00,.1065642152893932D-01,.4966599824214358D+00,
     &.1085488420737215D-01,.5180410345089337D+00,.1111843432827249D-01,
     &.5460483341177160D+00,.1149782035140485D-01,.5864065338958527D+00,
     &.3360562419718235D-02,.4078879630499024D-02,.7069329073407106D-02,
     &.4705262542208223D-02,.3332708647415080D-01,.5251963413756800D-02/
       data (recf(i),i=497,551)/                  .6738262760022461D-01,
     &.5730951315737387D-02,.1039837553534775D+00,.6152725177551053D-02,
     &.1405895253957461D+00,.6526132611753048D-02,.1759668677811218D+00,
     &.6858515620324401D-02,.2095479655628776D+00,.7155939019582245D-02,
     &.2411135586103733D+00,.7423415917481789D-02,.2706287495885053D+00,
     &.7665103884932816D-02,.2981548371545775D+00,.7884467592308588D-02,
     &.3238009173107087D+00,.8084411667986876D-02,.3476971265301733D+00,
     &.8267390411961632D-02,.3699801584879876D+00,.8435502610509800D-02,
     &.3907862310698117D+00,.8590582044778893D-02,.4102493117101150D+00,
     &.8734298913651078D-02,.4285042785031379D+00,.8868296093454296D-02,
     &.4456965009415761D+00,.8994399398069297D-02,.4620016328614911D+00,
     &.9114966064926033D-02,.4776628155317611D+00,.9233473962579847D-02,
     &.4930576673885005D+00,.9355505324217098D-02,.5088149715467647D+00,
     &.9490326792694050D-02,.5260104590494338D+00,.9653243364999375D-02,
     &.5464780715474396D+00,.9868600991484903D-02,.5732601989067468D+00,
     &.1017225716053114D-01,.6111363239215825D+00,.2950437825487400D-02,
     &.3581555112941687D-02,.7078620126622295D-02,.4132082643231012D-02,
     &.3337886298886408D-01,.4612740002118506D-02,.6750261702445300D-01,
     &.5034004128539568D-02,.1041911190752577D+00,.5405066674282486D-02/
       data (recf(i),i=552,606)/                  .1408978061674918D+00,
     &.5733674167146439D-02,.1763848697593524D+00,.6026255228800488D-02,
     &.2100805343046851D+00,.6288120524548417D-02,.2417622608971423D+00,
     &.6523660186046708D-02,.2713924596018528D+00,.6736515451660314D-02,
     &.2990302271326234D+00,.6929720582831246D-02,.3247828257866806D+00,
     &.7105817882799941D-02,.3487787217481978D+00,.7266950671260499D-02,
     &.3711526971215219D+00,.7414939513031372D-02,.3920379717842592D+00,
     &.7551347241996011D-02,.4115626674210771D+00,.7677539108440712D-02,
     &.4298493831220752D+00,.7794746313920410D-02,.4470176901244603D+00,
     &.7904145004284749D-02,.4631903418970273D+00,.8006969465604538D-02,
     &.4785052149162398D+00,.8104689159161024D-02,.4931367316839517D+00,
     &.8199295885823026D-02,.5073330960096394D+00,.8293770604485683D-02,
     &.5214794128662904D+00,.8392826136597125D-02,.5362017345344410D+00,
     &.8504035938797334D-02,.5525322098044688D+00,.8639408222270374D-02,
     &.5721561451757253D+00,.8817216208331321D-02,.5977434938944347D+00,
     &.9063189668758037D-02,.6332939739614603D+00,.2611205692460249D-02,
     &.3170101091465017D-02,.7086356419628467D-02,.3657766349841741D-02,
     &.3342206181054271D-01,.4083661809252030D-02,.6760295824860309D-01,
     &.4457037354538724D-02,.1043649740346678D+00,.4786009222605838D-02/
       data (recf(i),i=607,661)/                  .1411570048315963D+00,
     &.5077417957619811D-02,.1767374059456163D+00,.5336940476209487D-02,
     &.2105311844047106D+00,.5569266597573038D-02,.2423131187370251D+00,
     &.5778273412183262D-02,.2720433970782992D+00,.5967176862421978D-02,
     &.2997792878733318D+00,.6138656970674597D-02,.3256265499647623D+00,
     &.6294959076405515D-02,.3497123813254361D+00,.6437975105563999D-02,
     &.3721703830438909D+00,.6569309029799851D-02,.3931323996098995D+00,
     &.6690330403814164D-02,.4127244218331418D+00,.6802219694622686D-02,
     &.4310650456887266D+00,.6906009314473870D-02,.4482657882348953D+00,
     &.7002625080460826D-02,.4644331547605672D+00,.7092934548851317D-02,
     &.4796729107857300D+00,.7177811732959676D-02,.4940976831610963D+00,
     &.7258232689273568D-02,.5078399409574430D+00,.7335424050171227D-02,
     &.5210737517163408D+00,.7411097366633641D-02,.5340506436534539D+00,
     &.7487815663348852D-02,.5471575324915496D+00,.7569550927915151D-02,
     &.5610077596302204D+00,.7662488506257662D-02,.5765785117760304D+00,
     &.7776079157718089D-02,.5954049494922431D+00,.7924148666854938D-02,
     &.6198219542319461D+00,.8125409481841734D-02,.6531839888628567D+00,
     &.2329161903934607D-02,.2827942346666575D-02,.7092767801324573D-02,
     &.3263256653829519D-02,.3345792087848227D-01,.3643524065941897D-02/
       data (recf(i),i=662,716)/                  .6768640663901933D-01,
     &.3976979208186409D-02,.1045098632090649D+00,.4270848280167040D-02,
     &.1413735187729310D+00,.4531221910361700D-02,.1770326285691667D+00,
     &.4763154723278143D-02,.2109095929304028D+00,.4970822485283018D-02,
     &.2427770065864133D+00,.5157677387973698D-02,.2725932360325259D+00,
     &.5326583047275474D-02,.3004140385861841D+00,.5479926007282439D-02,
     &.3263439258620141D+00,.5619705807004115D-02,.3505090634425326D+00,
     &.5747607114086801D-02,.3730421698484556D+00,.5865057476337201D-02,
     &.3940742505245702D+00,.5973273864880296D-02,.4137303052977993D+00,
     &.6073300781079440D-02,.4321274134643955D+00,.6166042448892977D-02,
     &.4493743356584487D+00,.6252291618234036D-02,.4655722399143190D+00,
     &.6332757862946729D-02,.4808165109947856D+00,.6408099103723341D-02,
     &.4951999355274510D+00,.6478961616347051D-02,.5088179499753192D+00,
     &.6546036269964925D-02,.5217771724376283D+00,.6610142518258857D-02,
     &.5342092003559118D+00,.6672357066182766D-02,.5462927399000752D+00,
     &.6734211186973384D-02,.5582886150793852D+00,.6797988292026415D-02,
     &.5705940457094262D+00,.6867157346618580D-02,.5838243594874886D+00,
     &.6946966207628190D-02,.5989305413025341D+00,.7045163972244657D-02,
     &.6173557330297905D+00,.7172668498534273D-02,.6412138054673769D+00/
       data (recf(i),i=717,771)/                  .7343669551813224D-02,
     &.6734212283950197D+00,.2088962222558656D-02,.2536501141538391D-02,
     &.7098312096014856D-02,.2927174164471204D-02,.3348897507192802D-01,
     &.3268517695916878D-02,.6775879346915350D-01,.3567904939390041D-02,
     &.1046357814625728D+00,.3831806772332035D-02,.1415620734408565D+00,
     &.4065677361393518D-02,.1772903102475344D+00,.4274043151934446D-02,
     &.2112406912998520D+00,.4460643560808929D-02,.2431839618927255D+00,
     &.4628570053481958D-02,.2730769378663360D+00,.4780387068478273D-02,
     &.3009740811016449D+00,.4918231891978967D-02,.3269788261694700D+00,
     &.5043895306797768D-02,.3512164460435231D+00,.5158886125836192D-02,
     &.3738189139504120D+00,.5264482738771292D-02,.3949165867949340D+00,
     &.5361774417899160D-02,.4146338301275897D+00,.5451694687903237D-02,
     &.4330869551935628D+00,.5535048699155984D-02,.4503835468466499D+00,
     &.5612536310885108D-02,.4666226835980815D+00,.5684772523136652D-02,
     &.4818958265481784D+00,.5752307033652531D-02,.4962883637349025D+00,
     &.5815645095778566D-02,.5098819947919942D+00,.5875272604331699D-02,
     &.5227583713586083D+00,.5931689562795883D-02,.5350047139508995D+00,
     &.5985457948104635D-02,.5467225528673463D+00,.6037272660724469D-02,
     &.5580413411348073D+00,.6088067820748090D-02,.5691395094905580D+00/
       data (recf(i),i=772,826)/                  .6139174903693435D-02,
     &.5802765872262356D+00,.6192552890264469D-02,.5918411775133858D+00,
     &.6251110114702117D-02,.6044203697962170D+00,.6319123951993900D-02,
     &.6188952404877303D+00,.6402719127039078D-02,.6365612063366765D+00,
     &.6510255774185010D-02,.6592545134576846D+00,.6652267840556133D-02,
     &.6894261807467007D+00,.1884070525273232D-02,.2287865166522269D-02,
     &.7103061833161608D-02,.2640414555924009D-02,.3351561092037226D-01,
     &.2948504867845529D-02,.6782096720073962D-01,.3218776897022670D-02,
     &.1047441025101927D+00,.3457059211747976D-02,.1417245576749319D+00,
     &.3668264712145178D-02,.1775127832127863D+00,.3856470649104565D-02,
     &.2115271336150133D+00,.4025045325158119D-02,.2435368006995073D+00,
     &.4176773382240289D-02,.2734972927884814D+00,.4313964750721690D-02,
     &.3014619731636417D+00,.4438544635664399D-02,.3275333564649527D+00,
     &.4552126174118954D-02,.3518359485899692D+00,.4656068554335920D-02,
     &.3745010840861863D+00,.4751523397588311D-02,.3956585819539499D+00,
     &.4839471843717534D-02,.4154323324012205D+00,.4920754357241174D-02,
     &.4339381722015443D+00,.4996094895419632D-02,.4512831048238278D+00,
     &.5066120791624743D-02,.4675653273310193D+00,.5131379519337958D-02,
     &.4828747738428288D+00,.5192353422296487D-02,.4972940502928434D+00/
       data (recf(i),i=827,881)/                  .5249473538511721D-02,
     &.5108997625078284D+00,.5303133835275671D-02,.5237643599692746D+00,
     &.5353707549485549D-02,.5359587556465524D+00,.5401567951803124D-02,
     &.5475561617516223D+00,.5447116802332092D-02,.5586378279840697D+00,
     &.5490825128586838D-02,.5693017116602867D+00,.5533292803086976D-02,
     &.5796755768842047D+00,.5575335692614607D-02,.5899366290511978D+00,
     &.5618111542531525D-02,.6003405110877628D+00,.5663297103480300D-02,
     &.6112631579002907D+00,.5713326390293707D-02,.6232591473997167D+00,
     &.5771686951858539D-02,.6371386281547686D+00,.5843235308872718D-02,
     &.6540591245646711D+00,.5934415308321717D-02,.6756138475691030D+00,
     &.6053127455671619D-02,.7038679036965546D+00,.1708003175642380D-02,
     &.2074181532622547D-02,.7107154835723396D-02,.2393937473496849D-02,
     &.3353858755734360D-01,.2673414535955727D-02,.6787466315584890D-01,
     &.2918625868457997D-02,.1048377783441727D+00,.3134849526067874D-02,
     &.1418652820405583D+00,.3326534476511183D-02,.1777057753498972D+00,
     &.3497372950002969D-02,.2117760550049606D+00,.3650415148383601D-02,
     &.2438439995727389D+00,.3788182708532486D-02,.2738640090744446D+00,
     &.3912767381261045D-02,.3018885114907114D+00,.4025912539016479D-02,
     &.3280192335266340D+00,.4129078986309924D-02,.3523800195135267D+00/
       data (recf(i),i=882,936)/                  .4223497595500826D-02,
     &.3751016505461964D+00,.4310211297602812D-02,.3963134824357808D+00,
     &.4390108627214284D-02,.4161390116668366D+00,.4463950627129625D-02,
     &.4346937216368967D+00,.4532392561972495D-02,.4520842562952224D+00,
     &.4596001602217271D-02,.4684083677796514D+00,.4655271425546727D-02,
     &.4837553218910148D+00,.4710634539896850D-02,.4982065930947042D+00,
     &.4762473060977909D-02,.5118367810144925D+00,.4811128680431732D-02,
     &.5247147588413267D+00,.4856912649796014D-02,.5369051384586530D+00,
     &.4900116799220620D-02,.5484702216816450D+00,.4941026937249812D-02,
     &.5594727154468603D+00,.4979940477837468D-02,.5699796357324364D+00,
     &.5017190859382927D-02,.5800680268418966D+00,.5053182300865962D-02,
     &.5898333966076153D+00,.5088439691236800D-02,.5994021272105722D+00,
     &.5123679830879367D-02,.6089495614023648D+00,.5159911463507254D-02,
     &.6187259309040364D+00,.5198571556852845D-02,.6290926122189227D+00,
     &.5241701843449486D-02,.6405709220757223D+00,.5292158145073340D-02,
     &.6539038212088008D+00,.5353817484714892D-02,.6701255576222123D+00,
     &.5431693560278526D-02,.6906222214439113D+00,.5531783368314496D-02,
     &.7171434625046844D+00,.1486821222610860D-02,.1805710570607849D-02,
     &.7112315426865264D-02,.2084226678642758D-02,.3356758851849525D-01/
       data (recf(i),i=937,991)/                  .2327708403877407D-02,
     &.6794252207314784D-01,.2541382628281658D-02,.1049563283476954D+00,
     &.2729837665345000D-02,.1420436509870845D+00,.2896941194350830D-02,
     &.1779508117794810D+00,.3045903028705765D-02,.2120926886975343D+00,
     &.3179374802697815D-02,.2442355417695174D+00,.3299548613427395D-02,
     &.2743324022628396D+00,.3408242823717034D-02,.3024345386679296D+00,
     &.3506972940214846D-02,.3286426953033458D+00,.3597008842757047D-02,
     &.3530798855126308D+00,.3679420554768286D-02,.3758761896368122D+00,
     &.3755114750481509D-02,.3971603743636597D+00,.3824863905497256D-02,
     &.4170554403005484D+00,.3889329651040080D-02,.4356764468047173D+00,
     &.3949081575434511D-02,.4531296561693286D+00,.4004612453985609D-02,
     &.4695124355979618D+00,.4056350682593341D-02,.4849135873526999D+00,
     &.4104670535603563D-02,.4994139167345992D+00,.4149900758174290D-02,
     &.5130869342770428D+00,.4192331932996859D-02,.5259996460201358D+00,
     &.4232223028377952D-02,.5382134282701107D+00,.4269807540393116D-02,
     &.5497850203277219D+00,.4305299690448572D-02,.5607677074343013D+00,
     &.4338901239307418D-02,.5712128128876574D+00,.4370809641681970D-02,
     &.5811716792241682D+00,.4401228508048884D-02,.5906984004679025D+00,
     &.4430381681233268D-02,.5998536784140458D+00,.4458532691684085D-02/
       data (recf(i),i=992,1046)/                 .6087103236556506D+00,
     &.4486011931533010D-02,.6173611125584528D+00,.4513254549169507D-02,
     &.6259299431307766D+00,.4540852685066905D-02,.6345874840856066D+00,
     &.4569625913830092D-02,.6435727138959286D+00,.4600712887286172D-02,
     &.6532217324580875D+00,.4635683708228730D-02,.6640046346388578D+00,
     &.4676663849565939D-02,.6765693546213282D+00,.4726442410018883D-02,
     &.6917869917669843D+00,.4788505706770704D-02,.7107844374668939D+00,
     &.4866891449054713D-02,.7349355366345482D+00,.1306017463771605D-02,
     &.1586222301823758D-02,.7116548254584538D-02,.1830990793066982D-02,
     &.3359140173567671D-01,.2045006139498853D-02,.6799831246271334D-01,
     &.2232854278235691D-02,.1050539332243899D+00,.2398561178839865D-02,
     &.1421907384347976D+00,.2545520579058178D-02,.1781532257250431D+00,
     &.2676548965485182D-02,.2123547392925449D+00,.2793973015829898D-02,
     &.2445602455405584D+00,.2899716152387136D-02,.2747216803842120D+00,
     &.2995373840546507D-02,.3028893817005614D+00,.3082275794427197D-02,
     &.3291633014860622D+00,.3161536205810221D-02,.3536657782024154D+00,
     &.3234093916059993D-02,.3765263180987164D+00,.3300744456546394D-02,
     &.3978732008486888D+00,.3362165629048766D-02,.4178290151302299D+00,
     &.3418937992980868D-02,.4365084718269216D+00,.3471561346521851D-02/
       data (recf(i),i=1047,1101)/                .4540175351893392D+00,
     &.3520468055585971D-02,.4704533078136206D+00,.3566033899133432D-02,
     &.4859043358175601D+00,.3608586955892009D-02,.5004511375120722D+00,
     &.3648414949109787D-02,.5141668418466542D+00,.3685771386301219D-02,
     &.5271178746069831D+00,.3720880775408019D-02,.5393646641028532D+00,
     &.3753943164375919D-02,.5509623620882760D+00,.3785138236518451D-02,
     &.5619615952404358D+00,.3814629199517684D-02,.5724092815186362D+00,
     &.3842566733485300D-02,.5823495672538734D+00,.3869093316887355D-02,
     &.5918249678197956D+00,.3894348333795344D-02,.6008778302882713D+00,
     &.3918474488962536D-02,.6095522840022940D+00,.3941626226723230D-02,
     &.6178969082896937D+00,.3963981072925327D-02,.6259684294183701D+00,
     &.3985755098069792D-02,.6338368642128979D+00,.4007224021775936D-02,
     &.6415926549632834D+00,.4028751796394626D-02,.6493564801329609D+00,
     &.4050828701392520D-02,.6572925493481835D+00,.4074120789204574D-02,
     &.6656262309872097D+00,.4099531432075899D-02,.6746666723127143D+00,
     &.4128272797582476D-02,.6848343774930394D+00,.4161938830162795D-02,
     &.6966920189552035D+00,.4202559743668019D-02,.7109733150553114D+00,
     &.4252599514137185D-02,.7285986885865631D+00,.4314833901724070D-02,
     &.7506571441023483D+00,.1156343000755387D-02,.1404504257957441D-02/
       data (recf(i),i=1102,1156)/                .7120061440091347D-02,
     &.1621310545540372D-02,.3361118414788869D-01,.1810903712325119D-02,
     &.6804470745318069D-01,.1977340288814310D-02,.1051351961493040D+00,
     &.2124181703071103D-02,.1423133589772792D+00,.2254430165931555D-02,
     &.1783222117476625D+00,.2370577235911165D-02,.2125738546054181D+00,
     &.2474681149731246D-02,.2448322054772776D+00,.2568443387429520D-02,
     &.2750483122588739D+00,.2653275289328116D-02,.3032717571879833D+00,
     &.2730353093366796D-02,.3296018469202774D+00,.2800662377663822D-02,
     &.3541603675689819D+00,.2865033605362301D-02,.3770763523850850D+00,
     &.2924170474532069D-02,.3984776764701321D+00,.2978672551304409D-02,
     &.4184865833334669D+00,.3029053394756756D-02,.4372174902153924D+00,
     &.3075755134140899D-02,.4547761117084665D+00,.3119160251996799D-02,
     &.4712593365381469D+00,.3159601161354295D-02,.4867555224942895D+00,
     &.3197368036290039D-02,.5013450107382228D+00,.3232715256012342D-02,
     &.5151007425806035D+00,.3265866747398978D-02,.5280889117936882D+00,
     &.3297020454653708D-02,.5403696166527656D+00,.3326352123823280D-02,
     &.5519974959907685D+00,.3354018561785408D-02,.5630223474016425D+00,
     &.3380160512425799D-02,.5734897364461758D+00,.3404905286403203D-02,
     &.5834416154573392D+00,.3428369285319778D-02,.5929169809805325D+00/
       data (recf(i),i=1157,1211)/                .3450660577239604D-02,
     &.6019526114875673D+00,.3471881710106394D-02,.6105839432336187D+00,
     &.3492132995268978D-02,.6188461635645998D+00,.3511516558294448D-02,
     &.6267756293962285D+00,.3530141542180836D-02,.6344117559126738D+00,
     &.3548130962196775D-02,.6417995687010485D+00,.3565630852851213D-02,
     &.6489931729721409D+00,.3582822510857449D-02,.6560604660107810D+00,
     &.3599938804777734D-02,.6630894993854867D+00,.3617285646093928D-02,
     &.6701969730666828D+00,.3635269699036237D-02,.6775393849143700D+00,
     &.3654433055424882D-02,.6853273049469121D+00,.3675494570182433D-02,
     &.6938429782822218D+00,.3699395273559710D-02,.7034607793792287D+00,
     &.3727340922414418D-02,.7146686166221225D+00,.3760827388752240D-02,
     &.7280857713064263D+00,.3801623782166765D-02,.7444684055765861D+00,
     &.3851675608889073D-02,.7646881040691488D+00,.1031010001534118D-02,
     &.1252325311246044D-02,.7123010369617354D-02,.1445699377070109D-02,
     &.3362780174930066D-01,.1614821702453833D-02,.6808371387504373D-01,
     &.1763306316297053D-02,.1052035845952689D+00,.1894326321527065D-02,
     &.1424166653656900D+00,.2010556629704488D-02,.1784647514196653D+00,
     &.2114217175592618D-02,.2127589198858749D+00,.2207141797804957D-02,
     &.2450622274788554D+00,.2290846453223580D-02,.2753249929535957D+00/
       data (recf(i),i=1212,1266)/                .2366588574329418D-02,
     &.3035961777269406D+00,.2435416111636774D-02,.3299745572968218D+00,
     &.2498207136828419D-02,.3545814616992661D+00,.2555701517785104D-02,
     &.3775455317600242D+00,.2608526182258946D-02,.3989943047683779D+00,
     &.2657215287010062D-02,.4190497337604562D+00,.2702226368977643D-02,
     &.4378259868084232D+00,.2743953334055468D-02,.4554285653702798D+00,
     &.2782736954335406D-02,.4719541760694932D+00,.2818873396984337D-02,
     &.4874910203110806D+00,.2852621192436532D-02,.5021193021763500D+00,
     &.2884206960364913D-02,.5159118365654818D+00,.2913830143454304D-02,
     &.5289346889752990D+00,.2941666946802540D-02,.5412478085861132D+00,
     &.2967873641281116D-02,.5529056351590590D+00,.2992589359741792D-02,
     &.5639576722484762D+00,.3015938493650994D-02,.5744490272864715D+00,
     &.3038032783280771D-02,.5844209250482745D+00,.3058973186216636D-02,
     &.5939112061179864D+00,.3078851606347056D-02,.6029548271815797D+00,
     &.3097752568780461D-02,.6115843860580712D+00,.3115754935792624D-02,
     &.6198307020754208D+00,.3132933775825234D-02,.6277234924693847D+00,
     &.3149362522988406D-02,.6352921987752219D+00,.3165115600029473D-02,
     &.6425670346395947D+00,.3180271725033977D-02,.6495803491305275D+00,
     &.3194918182718492D-02,.6563684284758635D+00,.3209156415539186D-02/
       data (recf(i),i=1267,1321)/                .6629738949529550D+00,
     &.3223109375845204D-02,.6694489043071365D+00,.3236931170179888D-02,
     &.6758593906225592D+00,.3250819601641628D-02,.6822906541085610D+00,
     &.3265032236292007D-02,.6888546193197504D+00,.3279906508904962D-02,
     &.6956990815908923D+00,.3295884006098741D-02,.7030191567938097D+00,
     &.3313538195666039D-02,.7110708644516797D+00,.3333603169205194D-02,
     &.7201861618009667D+00,.3356997989054186D-02,.7307875953685270D+00,
     &.3384836571738381D-02,.7433987921789071D+00,.3418406731492194D-02,
     &.7586440923075351D+00,.3459095390889103D-02,.7772269308571649D+00,
     &.9250365544833106D-03,.1123642957047639D-02,.7125507954787448D-02,
     &.1297191729022396D-02,.3364188485222631D-01,.1448990578148682D-02,
     &.6811679517715986D-01,.1582279947362813D-02,.1052616326176499D+00,
     &.1699905063056458D-02,.1425044323407192D+00,.1804264479385807D-02,
     &.1785859728026249D+00,.1897348784865108D-02,.2129164803933608D+00,
     &.1980802350907045D-02,.2452582966079122D+00,.2055984495008354D-02,
     &.2755611351402767D+00,.2124022707017878D-02,.3038734422531415D+00,
     &.2185856628127736D-02,.3302935522984499D+00,.2242273566275012D-02,
     &.3549424157624133D+00,.2293936902368478D-02,.3779483458160068D+00,
     &.2341408747104221D-02,.3994385963604223D+00,.2385168029019629D-02/
       data (recf(i),i=1322,1376)/                .4195348752453025D+00,
     &.2425624979060079D-02,.4383511386477932D+00,.2463132778750725D-02,
     &.4559927053277270D+00,.2497996973377256D-02,.4725561248235422D+00,
     &.2530483119020393D-02,.4881294636884907D+00,.2560823028544199D-02,
     &.5027928098555960D+00,.2589219901343181D-02,.5166188766435952D+00,
     &.2615852559843530D-02,.5296736371549673D+00,.2640878968300298D-02,
     &.5420169498193963D+00,.2664439173073012D-02,.5537031542330661D+00,
     &.2686657775798838D-02,.5647816278134981D+00,.2707646029843779D-02,
     &.5752973009271112D+00,.2727503634730079D-02,.5852911327970665D+00,
     &.2746320291944519D-02,.5948005537838443D+00,.2764177077984035D-02,
     &.6038598823048999D+00,.2781147686318178D-02,.6125007272671443D+00,
     &.2797299589001256D-02,.6207523898606690D+00,.2812695171020565D-02,
     &.6286422822922934D+00,.2827392896389240D-02,.6361963859104227D+00,
     &.2841448574941446D-02,.6434397776056558D+00,.2854916813399088D-02,
     &.6503972618317222D+00,.2867852754331386D-02,.6570941566018360D+00,
     &.2880314232961717D-02,.6635572959395679D+00,.2892364515113282D-02,
     &.6698163290468718D+00,.2904075820167642D-02,.6759054182903871D+00,
     &.2915533879795195D-02,.6818654639717365D+00,.2926843832981620D-02,
     &.6877470127613112D+00,.2938137802322835D-02,.6936140357564597D+00/
       data (recf(i),i=1377,1431)/                .2949584518825127D-02,
     &.6995487849121559D+00,.2961401330692262D-02,.7056579402322907D+00,
     &.2973868789530239D-02,.7120802208967371D+00,.2987347661611534D-02,
     &.7189955103325420D+00,.3002297518260188D-02,.7266352712733982D+00,
     &.3019294818352010D-02,.7352935018498777D+00,.3039046378540213D-02,
     &.7453365750657752D+00,.3062391179964571D-02,.7572088751614536D+00,
     &.3090279763155639D-02,.7714291475982730D+00,.3123717004293759D-02,
     &.7885701646946724D+00,.8345540180276686D-03,.1013764009388037D-02,
     &.7127645881665573D-02,.1170376416196646D-02,.3365394644507796D-01,
     &.1307373442880687D-02,.6814514556190448D-01,.1427676957022642D-02,
     &.1053114143093287D+00,.1533852566639599D-02,.1425797600037836D+00,
     &.1628063138328361D-02,.1786901034175123D+00,.1712103666468225D-02,
     &.2130519545476170D+00,.1787456941371129D-02,.2454270534974888D+00,
     &.1855348698944282D-02,.2757646057348834D+00,.1916795617374952D-02,
     &.3041126250272165D+00,.1972644976800254D-02,.3305690752776119D+00,
     &.2023606687518786D-02,.3552545878917464D+00,.2070278907560604D-02,
     &.3782971999927905D+00,.2113168475544262D-02,.3998239259305988D+00,
     &.2152707223367783D-02,.4199562653334393D+00,.2189265039122685D-02,
     &.4388079933534241D+00,.2223160371923681D-02,.4564842715530542D+00/
       data (recf(i),i=1432,1486)/                .2254668720923272D-02,
     &.4730815132757386D+00,.2284029531207622D-02,.4886876674022503D+00,
     &.2311451825653930D-02,.5033827203854437D+00,.2337118829323729D-02,
     &.5172392978550550D+00,.2361191787080918D-02,.5303232962672809D+00,
     &.2383813132113577D-02,.5426945049786824D+00,.2405109129935129D-02,
     &.5544071973678512D+00,.2425192096951206D-02,.5655106807683926D+00,
     &.2444162273064263D-02,.5760498017810869D+00,.2462109412745146D-02,
     &.5860654077087350D+00,.2479114147565534D-02,.5955947674756841D+00,
     &.2495249164658244D-02,.6046719571441241D+00,.2510580239462893D-02,
     &.6133282164799120D+00,.2525167157099261D-02,.6215922842757839D+00,
     &.2539064554606323D-02,.6294907215579771D+00,.2552322716028421D-02,
     &.6370482335961646D+00,.2564988353967443D-02,.6442880040095037D+00,
     &.2577105414898926D-02,.6512320574262512D+00,.2588715951514404D-02,
     &.6579016713460415D+00,.2599861113932935D-02,.6643178633382018D+00,
     &.2610582323220381D-02,.6705019867831056D+00,.2620922705689986D-02,
     &.6764764773431096D+00,.2630928885293276D-02,.6822658035386095D+00,
     &.2640653254172355D-02,.6878976884216645D+00,.2650156867702621D-02,
     &.6934046853762110D+00,.2659513138546438D-02,.6988262090327137D+00,
     &.2668812530675645D-02,.7042111407137301D+00,.2678168471535181D-02/
       data (recf(i),i=1487,1541)/                .7096211435207562D+00,
     &.2687724694622147D-02,.7151348288632315D+00,.2697664171457851D-02,
     &.7208529025232651D+00,.2708219651065415D-02,.7269043646867752D+00,
     &.2719685534125355D-02,.7334537129952923D+00,.2732430278157871D-02,
     &.7407088522578688D+00,.2746907645451310D-02,.7489289818382320D+00,
     &.2763663754705104D-02,.7584310317783358D+00,.2783335043530471D-02,
     &.7695921849731747D+00,.2806630086397274D-02,.7828446706602086D+00,
     &.2834286424739216D-02,.7986575762535248D+00,.7568003719318281D-03,
     &.9193373558789463D-03,.7129483255703461D-02,.1061389190056637D-02,
     &.3366431715745472D-01,.1185658720812506D-02,.6816953450862767D-01,
     &.1294794350315917D-02,.1053542656584974D+00,.1391121645368222D-02,
     &.1426446444034435D+00,.1476601139740284D-02,.1787798638427945D+00,
     &.1552859914551620D-02,.2131688271751633D+00,.1621242048290884D-02,
     &.2455727658693272D+00,.1682858604168091D-02,.2759404560655418D+00,
     &.1738631136850422D-02,.3043195464558018D+00,.1789327643414803D-02,
     &.3308076883028777D+00,.1835591597189766D-02,.3555252435875009D+00,
     &.1877965170693660D-02,.3786000160301540D+00,.1916907758736738D-02,
     &.4001588169674569D+00,.1952810766576543D-02,.4203229689645817D+00,
     &.1986009452068925D-02,.4392060925299616D+00,.2016792448794289D-02/
       data (recf(i),i=1542,1596)/                .4569132141578328D+00,
     &.2045409461693157D-02,.4735406293653057D+00,.2072077518345742D-02,
     &.4891761844801786D+00,.2096986074148179D-02,.5038997769424400D+00,
     &.2120301203878496D-02,.5177839552843484D+00,.2142169061433792D-02,
     &.5308945491240521D+00,.2162718750455528D-02,.5432912893815535D+00,
     &.2182064718440740D-02,.5550283971155782D+00,.2200308763672422D-02,
     &.5661551304354499D+00,.2217541726297687D-02,.5767162856210821D+00,
     &.2233844920932215D-02,.5867526525797927D+00,.2249291357368938D-02,
     &.5963014271336025D+00,.2263946787643312D-02,.6053965840290423D+00,
     &.2277870611359356D-02,.6140692154201157D+00,.2291116666449641D-02,
     &.6223478401763524D+00,.2303733929177061D-02,.6302586899076299D+00,
     &.2315767145022433D-02,.6378259782215101D+00,.2327257411049353D-02,
     &.6450721605572670D+00,.2338242730369626D-02,.6520181930839076D+00,
     &.2348758560479908D-02,.6586838007193958D+00,.2358838379587591D-02,
     &.6650877664469207D+00,.2368514298725632D-02,.6712482569089239D+00,
     &.2377817752650840D-02,.6771832029067319D+00,.2386780309439892D-02,
     &.6829107580966474D+00,.2395434647564189D-02,.6884498650357838D+00,
     &.2403815760225585D-02,.6938209649667989D+00,.2411962459936224D-02,
     &.6990468964641347D+00,.2419919271518965D-02,.7041540382931724D+00/
       data (recf(i),i=1597,1651)/                .2427738818137931D-02,
     &.7091737632918418D+00,.2435484820891955D-02,.7141442820039481D+00,
     &.2443235844427740D-02,.7191129655073966D+00,.2451089922557623D-02,
     &.7241392431779231D+00,.2459170177943326D-02,.7292981672564718D+00,
     &.2467631490367586D-02,.7346847122783368D+00,.2476668140634181D-02,
     &.7404188178940524D+00,.2486522120258960D-02,.7466510642206331D+00,
     &.2497491395393781D-02,.7535686551625321D+00,.2509936783254573D-02,
     &.7614010329283727D+00,.2524285190189721D-02,.7704239097236394D+00,
     &.2541025785320841D-02,.7809597539113882D+00,.2560694413103588D-02,
     &.7933718506147267D+00,.2583840650378180D-02,.8080481691101191D+00,
     &.6694334606038313D-03,.8132301052210061D-03,.7131535707087396D-02,
     &.9389134630471259D-03,.3367590696807291D-01,.1048872877008853D-02,
     &.6819680438492585D-01,.1145449798724469D-02,.1054022066039507D+00,
     &.1230700342275683D-02,.1427172823544694D+00,.1306357911890783D-02,
     &.1788804222099017D+00,.1373861096813421D-02,.2132998613259399D+00,
     &.1434398265869949D-02,.2457362721216975D+00,.1488951760223115D-02,
     &.2761379595225904D+00,.1538336359526358D-02,.3045521719476219D+00,
     &.1583231069003380D-02,.3310762180672609D+00,.1624204791259703D-02,
     &.3558301646036297D+00,.1661736860472097D-02,.3789415592809941D+00/
       data (recf(i),i=1652,1706)/                .1696233421123923D-02,
     &.4005369903859170D+00,.1728040504318892D-02,.4207375854097784D+00,
     &.1757454499192354D-02,.4396567938240963D+00,.1784730573761186D-02,
     &.4573994919708219D+00,.1810089479804955D-02,.4740618435452991D+00,
     &.1833723080526557D-02,.4897315792782505D+00,.1855798864682184D-02,
     &.5044884954562696D+00,.1876463652712977D-02,.5184050523373731D+00,
     &.1895846655550170D-02,.5315470026926565D+00,.1914062012192412D-02,
     &.5439740105721099D+00,.1931210905475803D-02,.5557402385626821D+00,
     &.1947383334819745D-02,.5668948928266588D+00,.1962659608717758D-02,
     &.5774827218315940D+00,.1977111607276341D-02,.5875444685962498D+00,
     &.1990803855373865D-02,.5971172785245406D+00,.2003794439402125D-02,
     &.6062350661340985D+00,.2016135794604195D-02,.6149288446198813D+00,
     &.2027875385387674D-02,.6232270224896007D+00,.2039056297414846D-02,
     &.6311556716380911D+00,.2049717757561867D-02,.6387387713071162D+00,
     &.2059895595863033D-02,.6459984324868040D+00,.2069622662222278D-02,
     &.6529551075171532D+00,.2078929209927277D-02,.6596277899957604D+00,
     &.2087843257818011D-02,.6660342106414446D+00,.2096390943345113D-02,
     &.6721910355553485D+00,.2104596879732886D-02,.6781140744196829D+00,
     &.2112484532090143D-02,.6838185076461196D+00,.2120076629662863D-02/
       data (recf(i),i=1707,1761)/                .6893191434072465D+00,
     &.2127395634587425D-02,.6946307179415828D+00,.2134464291583733D-02,
     &.6997682556089891D+00,.2141306288121780D-02,.7047475089834524D+00,
     &.2147947060772256D-02,.7095855038862377D+00,.2154414790705499D-02,
     &.7143012197281974D+00,.2160741639474206D-02,.7189164418059094D+00,
     &.2166965284859545D-02,.7234568290808686D+00,.2173130824735037D-02,
     &.7279532479715103D+00,.2179293122829263D-02,.7324434288116620D+00,
     &.2185519670818465D-02,.7369740050569291D+00,.2191894031123936D-02,
     &.7416029929155098D+00,.2198519895768168D-02,.7464027556828451D+00,
     &.2205525735899346D-02,.7514634645503588D+00,.2213069905688267D-02,
     &.7568970038541067D+00,.2221345878507691D-02,.7628411564127120D+00,
     &.2230587003169685D-02,.7694637214568848D+00,.2241069745958558D-02,
     &.7769659387231029D+00,.2253113820886843D-02,.7855841979737009D+00,
     &.2267076946734208D-02,.7955885073370143D+00,.2283341347108146D-02,
     &.8072756369544127D+00,.2302288836034036D-02,.8209544135936097D+00,
     &.5959608869219697D-03,.7239930216451945D-03,.7133279984105104D-02,
     &.8359051841424598D-03,.3368576099913345D-01,.9338233700778646D-03,
     &.6822000216330427D-01,.1019831142421348D-02,.1054430127732599D+00,
     &.1095758216195270D-02,.1427791507549994D+00,.1163147152914922D-02/
       data (recf(i),i=1762,1816)/                .1789661337538298D+00,
     &.1223278163696114D-02,.2134116378937676D+00,.1277208785600386D-02,
     &.2458758686973440D+00,.1325813205911649D-02,.2763067382696023D+00,
     &.1369816492784942D-02,.3047511618662466D+00,.1409822883247175D-02,
     &.3313061633300986D+00,.1446338629616970D-02,.3560915640558304D+00,
     &.1479790274040400D-02,.3792346990474591D+00,.1510539225034315D-02,
     &.4008619710852165D+00,.1538893395093506D-02,.4210943451273732D+00,
     &.1565116520065816D-02,.4400451277298846D+00,.1589435653602577D-02,
     &.4578190692609130D+00,.1612047223434259D-02,.4745122222302962D+00,
     &.1633121950929719D-02,.4902122192180130D+00,.1652808868600187D-02,
     &.5049987699443599D+00,.1671238618449212D-02,.5189442584592429D+00,
     &.1688526174140261D-02,.5321143706133679D+00,.1704773099175760D-02,
     &.5445687118458739D+00,.1720069429524861D-02,.5563613934916154D+00,
     &.1734495250750528D-02,.5675415768232118D+00,.1748122025406067D-02,
     &.5781539706502275D+00,.1761013715337100D-02,.5882392821853880D+00,
     &.1773227734809691D-02,.5978346230971775D+00,.1784815763538346D-02,
     &.6069738738486110D+00,.1795824443291854D-02,.6156880099806808D+00,
     &.1806295977494385D-02,.6240053941930317D+00,.1816268649874427D-02,
     &.6319520380671680D+00,.1825777275563891D-02,.6395518371737207D+00/
       data (recf(i),i=1817,1871)/                .1834853595977716D-02,
     &.6468267831733791D+00,.1843526627209035D-02,.6537971564070679D+00,
     &.1851822970482422D-02,.6604817024077319D+00,.1859767092366030D-02,
     &.6668977957794100D+00,.1867381581918830D-02,.6730615950017611D+00,
     &.1874687391723729D-02,.6789881919521959D+00,.1881704069826390D-02,
     &.6846917603173716D+00,.1888449989971118D-02,.6901857076183885D+00,
     &.1894942588218185D-02,.6954828363313554D+00,.1901198615070921D-02,
     &.7005955205839351D+00,.1907234413674205D-02,.7055359061912127D+00,
     &.1913066236513566D-02,.7103161434075709D+00,.1918710615393293D-02,
     &.7149486637645913D+00,.1924184802344570D-02,.7194465147858177D+00,
     &.1929507302533566D-02,.7238237692546146D+00,.1934698524185406D-02,
     &.7280960290730413D+00,.1939781574913619D-02,.7322810475484290D+00,
     &.1944783238404899D-02,.7363994980480986D+00,.1949735169676031D-02,
     &.7404759210790995D+00,.1954675350226750D-02,.7445398854214426D+00,
     &.1959649844884542D-02,.7486274009877383D+00,.1964714897577384D-02,
     &.7527826199581922D+00,.1969939389932244D-02,.7570598558183508D+00,
     &.1975407658858821D-02,.7615259331570643D+00,.1981222619086292D-02,
     &.7662628484770114D+00,.1987509053119538D-02,.7713706654164001D+00,
     &.1994416800895364D-02,.7769704756416124D+00,.2002123390500117D-02/
       data (recf(i),i=1872,1926)/                .7832071162571111D+00,
     &.2010835389853712D-02,.7902511335384258D+00,.2020787431823991D-02,
     &.7982992156041412D+00,.2032237508156190D-02,.8075719965967809D+00,
     &.2045456833973120D-02,.8183078136932541D+00,.2060712530690398D-02,
     &.8307507916355182D+00,.5338928457208802D-03,.6486041055927330D-03,
     &.7134757145335637D-02,.7488783022520046D-03,.3369410911427631D-01,
     &.8366191792139075D-03,.6823966328875076D-01,.9136926200367255D-03,
     &.1054776147161770D+00,.9817372289705743D-03,.1428316413133060D+00,
     &.1042134467481678D-02,.1790388973189992D+00,.1096030851195136D-02,
     &.2135065916571426D+00,.1144373475917349D-02,.2459945403613866D+00,
     &.1187945232547193D-02,.2764503284138904D+00,.1227395462237520D-02,
     &.3049205940601908D+00,.1263265290388587D-02,.3315021246228428D+00,
     &.1296008087702652D-02,.3563145379544189D+00,.1326005836926910D-02,
     &.3794849926333959D+00,.1353582187798311D-02,.4011397375442878D+00,
     &.1379012879924876D-02,.4213996026736057D+00,.1402534089465843D-02,
     &.4403777756781442D+00,.1424349141395722D-02,.4581789018294929D+00,
     &.1444633933714447D-02,.4748989405223802D+00,.1463541343583908D-02,
     &.4906254417290438D+00,.1481204825549461D-02,.5054380418632013D+00,
     &.1497741365648968D-02,.5194090599684319D+00,.1513253919449596D-02/
       data (recf(i),i=1927,1981)/                .5326041243428160D+00,
     &.1527833434485045D-02,.5450827895910497D+00,.1541560536285142D-02,
     &.5568991222679660D+00,.1554506940716115D-02,.5681022442899654D+00,
     &.1566736642551277D-02,.5787368298952323D+00,.1578306920207553D-02,
     &.5888435558143096D+00,.1589269188760702D-02,.5984595065104622D+00,
     &.1599669727196656D-02,.6076185375122531D+00,.1609550300992578D-02,
     &.6163516003945891D+00,.1618948697263611D-02,.6246870331241742D+00,
     &.1627899186642147D-02,.6326508194315294D+00,.1636432923609736D-02,
     &.6402668207058439D+00,.1644578295050302D-02,.6475569836949800D+00,
     &.1652361225240263D-02,.6545415270716049D+00,.1659805444262448D-02,
     &.6612391097238539D+00,.1666932725870720D-02,.6676669834625878D+00,
     &.1673763100099416D-02,.6738411327196651D+00,.1680315045376260D-02,
     &.6797764037527307D+00,.1686605664538830D-02,.6854866258810527D+00,
     &.1692650848960719D-02,.6909847273635962D+00,.1698465434958682D-02,
     &.6962828487057689D+00,.1704063356777271D-02,.7013924564580136D+00,
     &.1709457800739110D-02,.7063244609630222D+00,.1714661365618782D-02,
     &.7110893420369401D+00,.1719686234962450D-02,.7156972872545727D+00,
     &.1724544367953908D-02,.7201583483730915D+00,.1729247716543003D-02,
     &.7244826224989144D+00,.1733808477926433D-02,.7286804659046311D+00/
       data (recf(i),i=1982,2036)/                .1738239393121469D-02,
     &.7327627499609472D+00,.1742554104306060D-02,.7367411704788124D+00,
     &.1746767585796486D-02,.7406286238588718D+00,.1750896665936375D-02,
     &.7444396657879347D+00,.1754960659646605D-02,.7481910707203003D+00,
     &.1758982133685755D-02,.7519025128611727D+00,.1762987828362212D-02,
     &.7555973915123057D+00,.1767009759806471D-02,.7593038249053531D+00,
     &.1771086524820975D-02,.7630558361569231D+00,.1775264824037991D-02,
     &.7668947513550588D+00,.1779601206067178D-02,.7708708209501843D+00,
     &.1784164011861242D-02,.7750450585439046D+00,.1789035459753730D-02,
     &.7794912616154927D+00,.1794313751342973D-02,.7842981311142257D+00,
     &.1800114989664162D-02,.7895713343902400D+00,.1806574577713146D-02,
     &.7954352514919308D+00,.1813847605191753D-02,.8020340030828954D+00,
     &.1822107542131273D-02,.8095311799649906D+00,.1831542366827648D-02,
     &.8181074940047334D+00,.1842347120393410D-02,.8279553875468303D+00,
     &.1854711901010908D-02,.8392695503623823D+00,.4691372119147557D-03,
     &.5699474625461372D-03,.7136296792849224D-02,.6580754336890405D-03,
     &.3370281338535284D-01,.7351932064547252D-03,.6826017144205968D-01,
     &.8029397345236971D-03,.1055137237879695D+00,.8627543786402343D-03,
     &.1428864460309730D+00,.9158506408206921D-03,.1791149115084724D+00/
       data (recf(i),i=2037,2091)/                .9632356479495138D-03,
     &.2136058483079311D+00,.1005741352235223D-02,.2461186723532833D+00,
     &.1044055462224991D-02,.2766006332016539D+00,.1078748370226813D-02,
     &.3050980855476346D+00,.1110295406319466D-02,.3317075749743232D+00,
     &.1139094812215725D-02,.3565485115490737D+00,.1165482118715345D-02,
     &.3797478738863422D+00,.1189741614017051D-02,.4014317539275095D+00,
     &.1212115499934443D-02,.4217208439626945D+00,.1232811224277848D-02,
     &.4407282102151493D+00,.1252007377470389D-02,.4585583904118685D+00,
     &.1269858457651580D-02,.4753072483903950D+00,.1286498741428814D-02,
     &.4910622490186240D+00,.1302045444889013D-02,.5059029528099534D+00,
     &.1316601318761973D-02,.5199016110853500D+00,.1330256790211312D-02,
     &.5331237917441728D+00,.1343091739511245D-02,.5456289955944984D+00,
     &.1355176981171679D-02,.5574712413729396D+00,.1366575504600436D-02,
     &.5686996086005114D+00,.1377343518145096D-02,.5793587340271479D+00,
     &.1387531331581560D-02,.5894892612975623D+00,.1397184105238413D-02,
     &.5991282456659598D+00,.1406342488530390D-02,.6083095167442896D+00,
     &.1415043166390284D-02,.6170640027934311D+00,.1423319328684793D-02,
     &.6254200202131955D+00,.1431201074983781D-02,.6334035318149531D+00,
     &.1438715764876940D-02,.6410383772701038D+00,.1445888322283529D-02/
       data (recf(i),i=2092,2146)/                .6483464788808477D+00,
     &.1452741500792328D-02,.6553480255553266D+00,.1459296115932730D-02,
     &.6620616376107076D+00,.1465571249361709D-02,.6685045147893643D+00,
     &.1471584429214883D-02,.6746925696634154D+00,.1477351790282003D-02,
     &.6806405484266425D+00,.1482888217204235D-02,.6863621409335069D+00,
     &.1488207473534721D-02,.6918700817451654D+00,.1493322319242184D-02,
     &.6971762438844605D+00,.1498244619060822D-02,.7022917269886535D+00,
     &.1502985443992799D-02,.7072269415837240D+00,.1507555168249840D-02,
     &.7119916912919050D+00,.1511963563977790D-02,.7165952549304837D+00,
     &.1516219896245011D-02,.7210464706718194D+00,.1520333020996940D-02,
     &.7253538247205133D+00,.1524311488991586D-02,.7295255473336399D+00,
     &.1528163659142506D-02,.7335697194752111D+00,.1531897825216204D-02,
     &.7374943939689233D+00,.1535522360469415D-02,.7413077357064542D+00,
     &.1539045885576599D-02,.7450181862940887D+00,.1542477466093639D-02,
     &.7486346594874230D+00,.1545826846727079D-02,.7521667748753935D+00,
     &.1549104730812699D-02,.7556251385226099D+00,.1552323114613247D-02,
     &.7590216806350385D+00,.1555495687246740D-02,.7623700617184821D+00,
     &.1558638308120671D-02,.7656861600411088D+00,.1561769574454004D-02,
     &.7689886543012415D+00,.1564911491470015D-02,.7722997159327905D+00/
       data (recf(i),i=2147,2201)/                .1568090256606748D-02,
     &.7756458249739051D+00,.1571337165828147D-02,.7790587211550291D+00,
     &.1574689643692269D-02,.7825764967678216D+00,.1578192387667105D-02,
     &.7862448284470696D+00,.1581898599185298D-02,.7901183291700260D+00,
     &.1585871246475873D-02,.7942619768635446D+00,.1590184264311382D-02,
     &.7987525387422443D+00,.1594923540574454D-02,.8036798573342322D+00,
     &.1600187467232215D-02,.8091477920349469D+00,.1606086745205357D-02,
     &.8152745179862013D+00,.1612743036112452D-02,.8221917759509073D+00,
     &.1620285966641179D-02,.8300425555166407D+00,.1628847946012576D-02,
     &.8389766067695471D+00,.1638556304338204D-02,.8491431596995497D+00,
     &.4152983993727092D-03,.5045486166497230D-03,.7137584516676377D-02,
     &.5825747502812775D-03,.3371009582559231D-01,.6508564339714855D-03,
     &.6827733616228911D-01,.7108440797559780D-03,.1055439591076286D+00,
     &.7638115008710665D-03,.1429323580617065D+00,.8108326421179146D-03,
     &.1791786256178479D+00,.8527988118498596D-03,.2136890926697665D+00,
     &.8904462894848839D-03,.2462228452780043D+00,.9243836974156776D-03,
     &.2767268570137912D+00,.9551158310881699D-03,.3052472497196298D+00,
     &.9830633537396637D-03,.3318803708195339D+00,.1008578703487142D-02,
     &.3567454603219721D+00,.1031958817701520D-02,.3799693497349148D+00/
       data (recf(i),i=2202,2256)/                .1053455282818112D-02,
     &.4016780028151404D+00,.1073282437929197D-02,.4219919994202545D+00,
     &.1091623864253886D-02,.4410243065866673D+00,.1108637603920062D-02,
     &.4588793741064279D+00,.1124460377322258D-02,.4756529875379346D+00,
     &.1139211008944395D-02,.4914325418394411D+00,.1152993225032100D-02,
     &.5062975349402129D+00,.1165897950461469D-02,.5203201620487163D+00,
     &.1178005204347923D-02,.5335659407189512D+00,.1189385672505494D-02,
     &.5460943265934672D+00,.1200102018320324D-02,.5579592979271204D+00,
     &.1210209980791963D-02,.5692098980166754D+00,.1219759298541784D-02,
     &.5798907312704116D+00,.1228794490820103D-02,.5900424125337194D+00,
     &.1237355520454566D-02,.5997019714820298D+00,.1245478358886877D-02,
     &.6089032150489527D+00,.1253195469650359D-02,.6176770513802826D+00,
     &.1260536223624142D-02,.6260517789476357D+00,.1267527256990541D-02,
     &.6340533443783142D+00,.1274192780889601D-02,.6417055723602077D+00,
     &.1280554850208162D-02,.6490303707240170D+00,.1286633597682126D-02,
     &.6560479135277237D+00,.1292447438469551D-02,.6627768046927827D+00,
     &.1298013249521533D-02,.6692342244811717D+00,.1303346527400884D-02,
     &.6754360608644642D+00,.1308461527646537D-02,.6813970276241349D+00,
     &.1313371388331951D-02,.6871307708378881D+00,.1318088240100810D-02/
       data (recf(i),i=2257,2311)/                .6926499652502783D+00,
     &.1322623304669488D-02,.6979664018971263D+00,.1326986983552337D-02,
     &.7030910682519399D+00,.1331188938584919D-02,.7080342220886105D+00,
     &.1335238165685898D-02,.7128054602082953D+00,.1339143063206245D-02,
     &.7174137831602817D+00,.1342911496162030D-02,.7218676570980016D+00,
     &.1346550857633121D-02,.7261750739540748D+00,.1350068128634278D-02,
     &.7303436111948426D+00,.1353469937828520D-02,.7343804925285762D+00,
     &.1356762622556962D-02,.7382926510963945D+00,.1359952292807390D-02,
     &.7420867968756576D+00,.1363044899939134D-02,.7457694902776256D+00,
     &.1366046312228174D-02,.7493472242304510D+00,.1368962399598178D-02,
     &.7528265174113487D+00,.1371799130263930D-02,.7562140217340938D+00,
     &.1374562682436201D-02,.7595166477149035D+00,.1377259574721566D-02,
     &.7627417119340626D+00,.1379896819392827D-02,.7658971114809663D+00,
     &.1382482103293908D-02,.7689915310079312D+00,.1385024001753947D-02,
     &.7720346888027838D+00,.1387532231477270D-02,.7750376290830836D+00,
     &.1390017948880056D-02,.7780130684490018D+00,.1392494100651911D-02,
     &.7809758049987551D+00,.1394975833264466D-02,.7839431988406547D+00,
     &.1397480967482131D-02,.7869357323732872D+00,.1400030542294373D-02,
     &.7899776573739077D+00,.1402649429581176D-02,.7930977330965135D+00/
       data (recf(i),i=2312,2366)/                .1405367015557528D-02,
     &.7963300544867956D+00,.1408217936717937D-02,.7997149612640727D+00,
     &.1411242845492577D-02,.8033000057004371D+00,.1414489162825236D-02,
     &.8071409378596117D+00,.1418011750045248D-02,.8113026400715047D+00,
     &.1421873399672547D-02,.8158599058260445D+00,.1426145003995468D-02,
     &.8208979110240450D+00,.1430905213100405D-02,.8265121682023010D+00,
     &.1436239345537763D-02,.8328076907174900D+00,.1442237275146430D-02,
     &.8398970330439564D+00,.1448990003811190D-02,.8478968323160648D+00,
     &.1456584666833031D-02,.8569224819892680D+00,.3704312646343119D-03,
     &.4500459652486567D-03,.7138650748115411D-02,.5196512293102546D-03,
     &.3371612727399277D-01,.5805664280060180D-03,.6829155668533947D-01,
     &.6340849446568049D-03,.1055690170651003D+00,.6813427081724704D-03,
     &.1429704232383565D+00,.7232974239492588D-03,.1792314731911011D+00,
     &.7607439326118729D-03,.2137581723047245D+00,.7943388257980795D-03,
     &.2463093367632228D+00,.8246248539513288D-03,.2768317147355182D+00,
     &.8520521766292996D-03,.3053712382217758D+00,.8769959255269664D-03,
     &.3320240933589542D+00,.8997703900142133D-03,.3569093819512950D+00,
     &.9206403645220593D-03,.3801538161111002D+00,.9398302000432457D-03,
     &.4018832556603875D+00,.9575310308788060D-03,.4222181893493448D+00/
       data (recf(i),i=2367,2421)/                .9739065619456522D-03,
     &.4412715038513685D+00,.9890977229041295D-03,.4591475776827830D+00,
     &.1003226429225417D-02,.4759421328870969D+00,.1016398637371945D-02,
     &.4917425076000317D+00,.1028706839793818D-02,.5066281487544273D+00,
     &.1040232113307033D-02,.5206712056814181D+00,.1051045809625145D-02,
     &.5339371545956020D+00,.1061210957702222D-02,.5464854138565951D+00,
     &.1070783432788909D-02,.5583699280903157D+00,.1079812935679248D-02,
     &.5696397102784545D+00,.1088343816748556D-02,.5803393375366302D+00,
     &.1096415772455089D-02,.5905094001856708D+00,.1104064436547561D-02,
     &.6001869059172129D+00,.1111321883943327D-02,.6094056420120871D+00,
     &.1118217061857135D-02,.6181964990928702D+00,.1124776160068879D-02,
     &.6265877600346048D+00,.1131022930068676D-02,.6346053575793632D+00,
     &.1136978961092068D-02,.6422731040006305D+00,.1142663919667303D-02,
     &.6496128959041625D+00,.1148095758170784D-02,.6566448969707471D+00,
     &.1153290896971796D-02,.6633877011656901D+00,.1158264384001310D-02,
     &.6698584786727629D+00,.1163030034968892D-02,.6760731065636836D+00,
     &.1167600556950560D-02,.6820462859911614D+00,.1171987657658222D-02,
     &.6877916474950851D+00,.1176202142362038D-02,.6933218458373045D+00,
     &.1180254000157654D-02,.6986486456296808D+00,.1184152481040773D-02/
       data (recf(i),i=2422,2476)/                .7037829988914664D+00,
     &.1187906165063826D-02,.7087351155643580D+00,.1191523024697338D-02,
     &.7135145279256490D+00,.1195010481397003D-02,.7181301497708876D+00,
     &.1198375457282797D-02,.7225903311867283D+00,.1201624422765785D-02,
     &.7269029097019594D+00,.1204763440909706D-02,.7310752585900764D+00,
     &.1207798209286642D-02,.7351143331006848D+00,.1210734100078424D-02,
     &.7390267154202931D+00,.1213576199187821D-02,.7428186592069160D+00,
     &.1216329345156313D-02,.7464961346090013D+00,.1218998168739366D-02,
     &.7500648747695434D+00,.1221587134066625D-02,.7535304249333266D+00,
     &.1224100582415115D-02,.7568981954218305D+00,.1226542779750137D-02,
     &.7601735199195562D+00,.1228917969342930D-02,.7633617207306558D+00,
     &.1231230430958390D-02,.7664681829190014D+00,.1233484548321435D-02,
     &.7694984395410563D+00,.1235684886817594D-02,.7724582705209949D+00,
     &.1237836283660978D-02,.7753538181016055D+00,.1239943953067019D-02,
     &.7781917222296842D+00,.1242013609290138D-02,.7809792796931293D+00,
     &.1244051610712973D-02,.7837246313034256D+00,.1246065128479356D-02,
     &.7864369818849385D+00,.1248062343408890D-02,.7891268582478557D+00,
     &.1250052675057249D-02,.7918064106168962D+00,.1252047046705140D-02,
     &.7944897630608011D+00,.1254058189642781D-02,.7971934181681546D+00/
       data (recf(i),i=2477,2531)/                .1256100989185991D-02,
     &.7999367203284967D+00,.1258192873167162D-02,.8027423802037812D+00,
     &.1260354240858018D-02,.8056370599049799D+00,.1262608925972372D-02,
     &.8086520134810325D+00,.1264984681034516D-02,.8118237698928188D+00,
     &.1267513661363494D-02,.8151948348470072D+00,.1270232874562430D-02,
     &.8188143727609828D+00,.1273184545147458D-02,.8227388097836817D+00,
     &.1276416323542609D-02,.8270322725064919D+00,.1279981244520295D-02,
     &.8317667447053557D+00,.1283937313921567D-02,.8370217873957335D+00,
     &.1288346577738765D-02,.8428836291187909D+00,.1293273510746542D-02,
     &.8494434005718798D+00,.1298782562596834D-02,.8567942719004841D+00,
     &.1304934730786103D-02,.8650272688993226D+00,.3247324487808653D-03,
     &.3945315229438140D-03,.7139767532965116D-02,.4555578699667749D-03,
     &.3372244637254624D-01,.5089677216150194D-03,.6830646000320411D-01,
     &.5558945983563953D-03,.1055952873791690D+00,.5973340508520150D-03,
     &.1430103457525390D+00,.6341254407194094D-03,.1792869234600663D+00,
     &.6669653938797302D-03,.2138306885390802D+00,.6964293711881412D-03,
     &.2464001780084456D+00,.7229930587138452D-03,.2769419073380583D+00,
     &.7470509916283250D-03,.3055016129403092D+00,.7689319468807855D-03,
     &.3321753151236530D+00,.7889113759410096D-03,.3570819738779806D+00/
       data (recf(i),i=2532,2586)/                .8072213501811222D-03,
     &.3803481790268297D+00,.8240584940970029D-03,.4020996841123463D+00,
     &.8395903192560990D-03,.4224568848461668D+00,.8539602966613761D-03,
     &.4415325859147346D+00,.8672919359416600D-03,.4594310931563854D+00,
     &.8796920818161035D-03,.4762480638503880D+00,.8912535918781442D-03,
     &.4920707781580621D+00,.9020575233992923D-03,.5069786309167420D+00,
     &.9121749286872399D-03,.5210437244980383D+00,.9216683368023415D-03,
     &.5343314926832955D+00,.9305929826843983D-03,.5469013154214589D+00,
     &.9389978318087510D-03,.5588071025313237D+00,.9469264384779324D-03,
     &.5700978354396460D+00,.9544176680744464D-03,.5808180626619678D+00,
     &.9615063075283847D-03,.5910083486194518D+00,.9682235834936652D-03,
     &.6007056775836025D+00,.9745976039775797D-03,.6099438156993251D+00,
     &.9806537362012611D-03,.6187536345605122D+00,.9864149311092719D-03,
     &.6271633999553797D+00,.9919020030614374D-03,.6351990293204530D+00,
     &.9971338717269585D-03,.6428843212420017D+00,.1002127771980826D-02,
     &.6502411600835498D+00,.1006899436614548D-02,.6572896985355157D+00,
     &.1011463255869661D-02,.6640485206005014D+00,.1015832417146389D-02,
     &.6705347872580487D+00,.1020019027701906D-02,.6767643668025705D+00,
     &.1024034222710236D-02,.6827519516206730D+00,.1027888260690500D-02/
       data (recf(i),i=2587,2641)/                .6885111629699746D+00,
     &.1031590608007845D-02,.6940546451402459D+00,.1035150013900340D-02,
     &.6993941502179868D+00,.1038574577276322D-02,.7045406145358273D+00,
     &.1041871806352736D-02,.7095042277666803D+00,.1045048672060125D-02,
     &.7142944955177615D+00,.1048111656019274D-02,.7189202961898697D+00,
     &.1051066793794383D-02,.7233899327913630D+00,.1053919714044959D-02,
     &.7277111803329067D+00,.1056675674131018D-02,.7318913293773558D+00,
     &.1059339592671662D-02,.7359372262783432D+00,.1061916079514346D-02,
     &.7398553106107196D+00,.1064409463539858D-02,.7436516502756216D+00,
     &.1066823818705570D-02,.7473319747524817D+00,.1069162988716224D-02,
     &.7509017069698120D+00,.1071430610707160D-02,.7543659942763458D+00,
     &.1073630138329335D-02,.7577297390145481D+00,.1075764864638867D-02,
     &.7609976292302569D+00,.1077837945216411D-02,.7641741700960905D+00,
     &.1079852421973832D-02,.7672637166832609D+00,.1081811248148003D-02,
     &.7702705087876981D+00,.1083717315034678D-02,.7731987086032039D+00,
     &.1085573481080025D-02,.7760524421380459D+00,.1087382604024221D-02,
     &.7788358453933391D+00,.1089147576881027D-02,.7815531164629429D+00,
     &.1090871368639801D-02,.7842085748763442D+00,.1092557070691845D-02,
     &.7868067296884163D+00,.1094207950110233D-02,.7893523580223542D+00/
       data (recf(i),i=2642,2696)/                .1095827511049306D-02,
     &.7918505959922322D+00,.1097419565672728D-02,.7943070441648366D+00,
     &.1098988316160812D-02,.7967278899584393D+00,.1100538449478067D-02,
     &.7991200496055508D+00,.1102075246684770D-02,.8014913325065773D+00,
     &.1103604708628230D-02,.8038506309404848D+00,.1105133699816469D-02,
     &.8062081381315597D+00,.1106670112111886D-02,.8085755975333251D+00,
     &.1108223049519360D-02,.8109665857911905D+00,.1109803034693859D-02,
     &.8133968310607946D+00,.1111422236740111D-02,.8158845670237690D+00,
     &.1113094718271409D-02,.8184509208406880D+00,.1114836697350123D-02,
     &.8211203301382262D+00,.1116666816628120D-02,.8239209796093340D+00,
     &.1118606407495873D-02,.8268852415235039D+00,.1120679731089786D-02,
     &.8300500959832002D+00,.1122914170403457D-02,.8334574957353697D+00,
     &.1125340338439976D-02,.8371546264999228D+00,.1127992056534935D-02,
     &.8411939971620041D+00,.1130906145334969D-02,.8456332754097852D+00,
     &.1134121959793832D-02,.8505347650275864D+00,.1137680591275090D-02,
     &.8559644039843292D+00,.1141623657948933D-02,.8619901524142233D+00,
     &.1145991613898916D-02,.8686796434054859D+00,.2875557040695001D-03,
     &.3493682107340224D-03,.7140652840633392D-02,.4034136512140228D-03,
     &.3372745678580264D-01,.4507156263114678D-03,.6831827979235850D-01/
       data (recf(i),i=2697,2751)/                .4922776751319498D-03,
     &.1056161282055053D+00,.5289812044432154D-03,.1430420271934089D+00,
     &.5615693465755460D-03,.1793309426893349D+00,.5906588650957148D-03,
     &.2138882776015863D+00,.6167592500442515D-03,.2464723498787023D+00,
     &.6402916548621664D-03,.2770294926804611D+00,.6616053847329528D-03,
     &.3056052894425374D+00,.6809915244670573D-03,.3322956309926026D+00,
     &.6986939459404484D-03,.3572193667731841D+00,.7149181133975426D-03,
     &.3805029914283647D+00,.7298381072985964D-03,.4022721758944257D+00,
     &.7436022322429452D-03,.4226472436509502D+00,.7563375079342728D-03,
     &.4417409357850834D+00,.7681532808226864D-03,.4596575017826150D+00,
     &.7791441427437137D-03,.4764925487070491D+00,.7893923017930954D-03,
     &.4923333117464571D+00,.7989695184985524D-03,.5072591452826548D+00,
     &.8079386954131118D-03,.5213421151585618D+00,.8163551890154100D-03,
     &.5346476220690836D+00,.8242678979707145D-03,.5472350159192663D+00,
     &.8317201703560293D-03,.5591581791948586D+00,.8387505635876626D-03,
     &.5704660684234719D+00,.8453934839008121D-03,.5812032094226068D+00,
     &.8516797268548300D-03,.5914101459193340D+00,.8576369361234619D-03,
     &.6011238433265917D+00,.8632899945099273D-03,.6103780506208373D+00,
     &.8686613584996434D-03,.6192036237904337D+00,.8737713455743161D-03/
       data (recf(i),i=2752,2806)/                .6276288144678238D+00,
     &.8786383818419734D-03,.6356795272806593D+00,.8832792161976796D-03,
     &.6433795492571826D+00,.8877091061492647D-03,.6507507543610395D+00,
     &.8919419795673447D-03,.6578132859480015D+00,.8959905759071083D-03,
     &.6645857196542167D+00,.8998665698679513D-03,.6710852089553721D+00,
     &.9035806799802425D-03,.6773276153852716D+00,.9071427642160033D-03,
     &.6833276251737907D+00,.9105619043959713D-03,.6890988538586815D+00,
     &.9138464808966333D-03,.6946539402426604D+00,.9170042389371524D-03,
     &.7000046309052734D+00,.9200423475395102D-03,.7051618563365225D+00,
     &.9229674520990890D-03,.7101357996343394D+00,.9257857213720340D-03,
     &.7149359585989350D+00,.9285028895757936D-03,.7195712019621512D+00,
     &.9311242942067709D-03,.7240498204076717D+00,.9336549101012279D-03,
     &.7283795729669328D+00,.9360993802001778D-03,.7325677293146301D+00,
     &.9384620434241330D-03,.7366211084358048D+00,.9407469600177421D-03,
     &.7405461140927553D+00,.9429579346863432D-03,.7443487674837327D+00,
     &.9450985378153066D-03,.7480347374559778D+00,.9471721250379737D-03,
     &.7516093686126949D+00,.9491818553984218D-03,.7550777076367228D+00,
     &.9511307083407467D-03,.7584445281427596D+00,.9530214997467334D-03,
     &.7617143543649362D+00,.9548568972384669D-03,.7648914839873381D+00/
       data (recf(i),i=2807,2861)/                .9566394349615138D-03,
     &.7679800104318763D+00,.9583715280677518D-03,.7709838449309258D+00,
     &.9600554871248025D-03,.7739067387317117D+00,.9616935326914640D-03,
     &.7767523058059454D+00,.9632878103157287D-03,.7795240464721949D+00,
     &.9648404062341729D-03,.7822253723804877D+00,.9663533640789765D-03,
     &.7848596333593397D+00,.9678287029319074D-03,.7874301466854236D+00,
     &.9692684371035443D-03,.7899402294061247D+00,.9706745980610887D-03,
     &.7923932344258478D+00,.9720492589794330D-03,.7947925911586078D+00,
     &.9733945624476692D-03,.7971418516523020D+00,.9747127519265526D-03,
     &.7994447432038601D+00,.9760062076207416D-03,.8017052286081833D+00,
     &.9772774875013603D-03,.8039275753153862D+00,.9785293742870215D-03,
     &.8061164349067315D+00,.9797649292608988D-03,.8082769344340225D+00,
     &.9809875538617746D-03,.8104147812911904D+00,.9822010600294816D-03,
     &.8125363833872510D+00,.9834097502974072D-03,.8146489864477286D+00,
     &.9846185085894264D-03,.8167608302603168D+00,.9858329025720856D-03,
     &.8188813255629256D+00,.9870592982031829D-03,.8210212529979992D+00,
     &.9883049867630125D-03,.8231929850588618D+00,.9895783241000432D-03,
     &.8254107311438275D+00,.9908888810000608D-03,.8276908045991552D+00,
     &.9922476024127977D-03,.8300519088320196D+00,.9936669716440181D-03/
       data (recf(i),i=2862,2916)/                .8325154370399047D+00,
     &.9951611734340783D-03,.8351057766377819D+00,.9967462469844777D-03,
     &.8378506048585280D+00,.9984402163659223D-03,.8407811560526175D+00,
     &.1000263181294734D-02,.8439324337673380D+00,.1002237346041166D-02,
     &.8473433317092810D+00,.1004386958434913D-02,.8510566173745598D+00,
     &.1006738125009853D-02,.8551187210241215D+00,.1009318463085570D-02,
     &.8595792618934147D+00,.1012156547280817D-02,.8644902349222623D+00,
     &.1015281108386375D-02,.8699047776816512D+00,.1018719948988823D-02,
     &.8758754423894201D+00,.2516072353014572D-03,.3056959967091976D-03,
     &.7141513331248321D-02,.3529898095786411D-03,.3373232773731790D-01,
     &.3943840897337921D-03,.6832977329356830D-01,.4307567414106005D-03,
     &.1056363991444212D+00,.4628788405472162D-03,.1430728515532258D+00,
     &.4914005032824914D-03,.1793737852691615D+00,.5168612968898808D-03,
     &.2139443476569179D+00,.5397069428251232D-03,.2465426458452338D+00,
     &.5603058824394920D-03,.2771148378485354D+00,.5789637011504982D-03,
     &.3057063604036936D+00,.5959350504603019D-03,.3324129804044311D+00,
     &.6114332776489099D-03,.3573534416826857D+00,.6256381290223384D-03,
     &.3806541481236008D+00,.6387018947049087D-03,.4024406920346254D+00,
     &.6507543147370997D-03,.4228333282534617D+00,.6619065080162839D-03/
       data (recf(i),i=2917,2971)/                .4419447375155043D+00,
     &.6722541319677103D-03,.4598791158936434D+00,.6818799359437668D-03,
     &.4767320228947860D+00,.6908558354114982D-03,.4925906511342484D+00,
     &.6992446058387721D-03,.5075343166976332D+00,.7071012733743214D-03,
     &.5216350508314642D+00,.7144742625858456D-03,.5349582228611499D+00,
     &.7214063485450536D-03,.5475631541591855D+00,.7279354505312647D-03,
     &.5595037011924066D+00,.7340952968696949D-03,.5708287967136168D+00,
     &.7399159843938746D-03,.5815829447836500D+00,.7454244513185431D-03,
     &.5918066692004921D+00,.7506448786223955D-03,.6015369171138536D+00,
     &.7555990321360334D-03,.6108074207645846D+00,.7603065552321395D-03,
     &.6196490208139152D+00,.7647852201872223D-03,.6280899548719071D+00,
     &.7690511448239667D-03,.6361561147571614D+00,.7731189798710006D-03,
     &.6438712758203572D+00,.7770020715316172D-03,.6512573014043507D+00,
     &.7807126029873436D-03,.6583343252310160D+00,.7842617179394069D-03,
     &.6651209142221997D+00,.7876596287823588D-03,.6716342139918672D+00,
     &.7909157115868082D-03,.6778900789954702D+00,.7940385897245874D-03,
     &.6839031890937339D+00,.7970362076856718D-03,.6896871540821554D+00,
     &.7999158964005718D-03,.6952546075538913D+00,.8026844311858064D-03,
     &.7006172913010297D+00,.8053480832662807D-03,.7057861313157792D+00/
       data (recf(i),i=2972,3026)/                .8079126656911606D-03,
     &.7107713063270146D+00,.8103835743445347D-03,.7155823096970855D+00,
     &.8127658246549895D-03,.7202280054070732D+00,.8150640845261530D-03,
     &.7247166787741906D+00,.8172827039407797D-03,.7290560824713176D+00,
     &.8194257416320207D-03,.7332534783544617D+00,.8214969891654706D-03,
     &.7373156755481164D+00,.8234999927330391D-03,.7412490651900622D+00,
     &.8254380729235324D-03,.7450596521952805D+00,.8273143427041148D-03,
     &.7487530843625872D+00,.8291317238207998D-03,.7523346791167432D+00,
     &.8308929618041609D-03,.7558094481526250D+00,.8326006397480466D-03,
     &.7591821202261264D+00,.8342571910138194D-03,.7624571623184618D+00,
     &.8358649110001742D-03,.7656387993861767D+00,.8374259681086725D-03,
     &.7687310328982383D+00,.8389424140275449D-03,.7717376583539388D+00,
     &.8404161934509115D-03,.7746622819709052D+00,.8418491533472557D-03,
     &.7775083367312447D+00,.8432430518896748D-03,.7802790979757885D+00,
     &.8445995671611185D-03,.7829776987415892D+00,.8459203057504956D-03,
     &.7856071450464104D+00,.8472068113602371D-03,.7881703313360623D+00,
     &.8484605735526905D-03,.7906700563263103D+00,.8496830367716911D-03,
     &.7931090394909276D+00,.8508756097868982D-03,.7954899384715712D+00,
     &.8520396757221188D-03,.7978153677138051D+00,.8531766028449862D-03/
       data (recf(i),i=3027,3081)/                .8000879186670883D+00,
     &.8542877563141193D-03,.8023101819251651D+00,.8553745111013562D-03,
     &.8044847717273021D+00,.8564382663308516D-03,.8066143532903718D+00,
     &.8574804613037265D-03,.8087016734969329D+00,.8585025935063833D-03,
     &.8107495955250247D+00,.8595062389322027D-03,.8127611380708978D+00,
     &.8604930750795124D-03,.8147395198853895D+00,.8614649070224791D-03,
     &.8166882104165335D+00,.8624236969844256D-03,.8186109874227465D+00,
     &.8633715978728290D-03,.8205120024888089D+00,.8643109912587642D-03,
     &.8223958554353771D+00,.8652445302964463D-03,.8242676786542088D+00,
     &.8661751880747444D-03,.8261332324148928D+00,.8671063118639506D-03,
     &.8279990121599971D+00,.8680416836568219D-03,.8298723687144858D+00,
     &.8689855872887505D-03,.8317616421559534D+00,.8699428822395237D-03,
     &.8336763097907850D+00,.8709190839453324D-03,.8356271482143981D+00,
     &.8719204500557906D-03,.8376264087468918D+00,.8729540715222103D-03,
     &.8396880045621392D+00,.8740279666600877D-03,.8418277064895145D+00,
     &.8751511753462460D-03,.8440633426727417D+00,.8763338492436154D-03,
     &.8464149949224021D+00,.8775873323529088D-03,.8489051816016571D+00,
     &.8789242242430914D-03,.8515590131598368D+00,.8803584160127719D-03,
     &.8544043019385164D+00,.8819050864333951D-03,.8574716026611897D+00/
       data (recf(i),i=3082,3136)/                .8835806429514380D-03,
     &.8607941542532526D+00,.8854025895229178D-03,.8644076876951290D+00,
     &.8873893010135113D-03,.8683500591389794D+00,.8895596827024104D-03,
     &.8726606635323475D+00,.8919326940631082D-03,.8773795829320804D+00,
     &.8945267194288890D-03,.8825464274388153D+00,.2223192652822833D-03,
     &.2701145057190040D-03,.7142201961578917D-02,.3119065510375409D-03,
     &.3373622651253249D-01,.3484864321264339D-03,.6833897467046082D-01,
     &.3806297364820721D-03,.1056526311462410D+00,.4090176838012951D-03,
     &.1430975404471804D+00,.4342246342064952D-03,.1794081097583843D+00,
     &.4567272828059632D-03,.2139892833907438D+00,.4769194148963887D-03,
     &.2465990011834060D+00,.4951265406555741D-03,.2771832823761325D+00,
     &.5116186382371054D-03,.3057874476558851D+00,.5266206864730374D-03,
     &.3325071660393497D+00,.5403211724599088D-03,.3574610980955170D+00,
     &.5528788972227103D-03,.3807755762784305D+00,.5644284045399661D-03,
     &.4025761311147353D+00,.5750843154180597D-03,.4229829635932437D+00,
     &.5849447992701323D-03,.4421087071714556D+00,.5940943654599883D-03,
     &.4600575161312302D+00,.6026061192146514D-03,.4769249128064807D+00,
     &.6105435941597527D-03,.4927980565571165D+00,.6179622488629170D-03,
     &.5077562335628903D+00,.6249106954973403D-03,.5218714480507487D+00/
       data (recf(i),i=3137,3191)/                .6314317138679972D-03,
     &.5352090448328407D+00,.6375630925795459D-03,.5478283229614227D+00,
     &.6433383302749551D-03,.5597831185164269D+00,.6487872230220317D-03,
     &.5711223455807324D+00,.6539363586005573D-03,.5818904910810551D+00,
     &.6588095342876049D-03,.5921280630645866D+00,.6634281114812186D-03,
     &.6018719941845592D+00,.6678113179369536D-03,.6111560033299172D+00,
     &.6719765063612168D-03,.6200109188606474D+00,.6759393764906060D-03,
     &.6284649670553408D+00,.6797141664962577D-03,.6365440293006893D+00,
     &.6833138185165486D-03,.6442718713535327D+00,.6867501222863375D-03,
     &.6516703477465010D+00,.6900338401544602D-03,.6587595841259685D+00,
     &.6931748162308895D-03,.6655581400283747D+00,.6961820719554190D-03,
     &.6720831543307668D+00,.6990638900109890D-03,.6783504753604156D+00,
     &.7018278882011304D-03,.6843747774195197D+00,.7044810846600041D-03,
     &.6901696652750492D+00,.7070299555552817D-03,.6957477679800600D+00,
     &.7094804862707378D-03,.7011208232299629D+00,.7118382169105868D-03,
     &.7062997533135355D+00,.7141082828461975D-03,.7112947335920807D+00,
     &.7162954509237486D-03,.7161152543292093D+00,.7184041518653100D-03,
     &.7207701765965301D+00,.7204385093230375D-03,.7252677828954485D+00,
     &.7224023659844375D-03,.7296158230608452D+00,.7242993070741803D-03/
       data (recf(i),i=3192,3246)/                .7338215559473337D+00,
     &.7261326815532289D-03,.7378917873419190D+00,.7279056212778734D-03,
     &.7418329044971879D+00,.7296210583486120D-03,.7456509076357754D+00,
     &.7312817408508591D-03,.7493514387389932D+00,.7328902471654914D-03,
     &.7529398078995253D+00,.7344489990066897D-03,.7564210174894084D+00,
     &.7359602733269230D-03,.7597997843696403D+00,.7374262132138434D-03,
     &.7630805603462782D+00,.7388488378909974D-03,.7662675510594492D+00,
     &.7402300519233290D-03,.7693647334760027D+00,.7415716537192339D-03,
     &.7723758721433454D+00,.7428753434132293D-03,.7753045343511121D+00,
     &.7441427302069922D-03,.7781541043385911D+00,.7453753392414597D-03,
     &.7809277966791060D+00,.7465746180687906D-03,.7836286689677923D+00,
     &.7477419427901918D-03,.7862596339363163D+00,.7488786239238555D-03,
     &.7888234711170709D+00,.7499859120665148D-03,.7913228381802210D+00,
     &.7510650034123837D-03,.7937602820697135D+00,.7521170451945067D-03,
     &.7961382500690559D+00,.7531431411158165D-03,.7984591009343806D+00,
     &.7541443568405107D-03,.8007251162411515D+00,.7551217256207493D-03,
     &.8029385121019574D+00,.7560762541391865D-03,.8051014514262963D+00,
     &.7570089286545360D-03,.8072160569092491D+00,.7579207215452834D-03,
     &.8092844249546161D+00,.7588125983558522D-03,.8113086407595959D+00/
       data (recf(i),i=3247,3301)/                .7596855254600513D-03,
     &.8132907948125792D+00,.7605404784685073D-03,.8152330010832001D+00,
     &.7613784515200193D-03,.8171374172145180D+00,.7622004676113267D-03,
     &.8190062670610542D+00,.7630075901355534D-03,.8208418659532486D+00,
     &.7638009358163845D-03,.8226466491084072D+00,.7645816892425407D-03,
     &.8244232036498043D+00,.7653511192248359D-03,.8261743047383627D+00,
     &.7661105972153171D-03,.8279029563638637D+00,.7668616180436525D-03,
     &.8296124373828724D+00,.7676058232385866D-03,.8313063534256175D+00,
     &.7683450272099249D-03,.8329886953198786D+00,.7690812465664271D-03,
     &.8346639046910128D+00,.7698167328335732D-03,.8363369473861753D+00,
     &.7705540088076615D-03,.8380133953277569D+00,.7712959087329680D-03,
     &.8396995173132502D+00,.7720456224088709D-03,.8414023791296488D+00,
     &.7728067432140554D-03,.8431299531191393D+00,.7735833198629735D-03,
     &.8448912369930949D+00,.7743799114709090D-03,.8466963812112370D+00,
     &.7752016451809944D-03,.8485568235840266D+00,.7760542751796868D-03,
     &.8504854288745331D+00,.7769442413753351D-03,.8524966300217779D+00,
     &.7778787253163293D-03,.8546065661289764D+00,.7788657000623884D-03,
     &.8568332105076614D+00,.7799139696833306D-03,.8591964798032911D+00,
     &.7810331928461635D-03,.8617183125335791D+00,.7822338835880836D-03/
       data (recf(i),i=3302,3356)/                .8644227022726899D+00,
     &.7835273809189877D-03,.8673356673041829D+00,.7849257774604757D-03,
     &.8704851350330805D+00,.7864417960824488D-03,.8739007161188697D+00,
     &.7880886026971684D-03,.8776133406687649D+00,.7898795433580249D-03,
     &.8816547276255949D+00,.7918277950171212D-03,.8860566596308294D+00,
     &.7939459222137213D-03,.8908500402679119D+00,.1944104808172863D-03,
     &.2362079689279158D-03,.7142870645995927D-02,.2727565493010114D-03,
     &.3374001298187475D-01,.3047478195749429D-03,.6834791269705135D-01,
     &.3328598979745383D-03,.1056684019862271D+00,.3576884028469226D-03,
     &.1431215337265024D+00,.3797355245930193D-03,.1794414760981541D+00,
     &.3994180621642384D-03,.2140329776183974D+00,.4170803229573239D-03,
     &.2466538170423533D+00,.4330069180398755D-03,.2772498801896881D+00,
     &.4474339038008580D-03,.3058663763636773D+00,.4605579911155484D-03,
     &.3325988808094674D+00,.4725439839337465D-03,.3575659744772789D+00,
     &.4835307298323497D-03,.3808939216502081D+00,.4936358667461324D-03,
     &.4027081940104168D+00,.5029596128586884D-03,.4231289413573305D+00,
     &.5115878016710237D-03,.4422687522901076D+00,.5195943228286058D-03,
     &.4602317415082190D+00,.5270430946153614D-03,.4771133961932997D+00,
     &.5339896662637906D-03,.4930008443009192D+00,.5404825264865527D-03/
       data (recf(i),i=3357,3411)/                .5079733438010129D+00,
     &.5465641777837335D-03,.5221028734536996D+00,.5522720230787040D-03,
     &.5354547549768437D+00,.5576391012124826D-03,.5480882663954187D+00,
     &.5626947000884556D-03,.5600572245757691D+00,.5674648702683609D-03,
     &.5714105259898375D+00,.5719728571649507D-03,.5821926413793687D+00,
     &.5762394663436609D-03,.5924440638839293D+00,.5802833735974854D-03,
     &.6022017124008726D+00,.5841213892159104D-03,.6114992931083616D+00,
     &.5877686840933180D-03,.6203676226097062D+00,.5912389839103901D-03,
     &.6288349163028890D+00,.5945447364939478D-03,.6369270455027728D+00,
     &.5976972565551031D-03,.6446677666447656D+00,.6007068512753596D-03,
     &.6520789256394533D+00,.6035829296188047D-03,.6591806401656133D+00,
     &.6063340977673600D-03,.6659914624065441D+00,.6089682426829690D-03,
     &.6725285244645835D+00,.6114926054781716D-03,.6788076684377859D+00,
     &.6139138460109964D-03,.6848435629139552D+00,.6162380999006194D-03,
     &.6906498074313085D+00,.6184710289781289D-03,.6962390262713238D+00,
     &.6206178660351128D-03,.7016229527864589D+00,.6226834546060953D-03,
     &.7068125053216800D+00,.6246722844146502D-03,.7118178556622645D+00,
     &.6265885230237052D-03,.7166484908292960D+00,.6284360441552101D-03,
     &.7213132689469173D+00,.6302184530805945D-03,.7258204698201126D+00/
       data (recf(i),i=3412,3466)/                .6319391094293539D-03,
     &.7301778407871102D+00,.6336011477170803D-03,.7343926383451021D+00,
     &.6352074958549976D-03,.7384716659907177D+00,.6367608918694911D-03,
     &.7424213086665331D+00,.6382638990313429D-03,.7462475641609494D+00,
     &.6397189195696667D-03,.7499560717702578D+00,.6411282071242605D-03,
     &.7535521384979339D+00,.6424938780717482D-03,.7570407630365846D+00,
     &.6438179218450410D-03,.7604266577520003D+00,.6451022103519516D-03,
     &.7637142688659893D+00,.6463485065869407D-03,.7669077950147200D+00,
     &.6475584725197199D-03,.7700112043418378D+00,.6487336763355520D-03,
     &.7730282502703713D+00,.6498755990944187D-03,.7759624860841656D+00,
     &.6509856408696007D-03,.7788172784380622D+00,.6520651264205267D-03,
     &.7815958199061201D+00,.6531153104498718D-03,.7843011406686931D+00,
     &.6541373824907589D-03,.7869361194320170D+00,.6551324714664429D-03,
     &.7895034936680300D+00,.6561016499619955D-03,.7920058692573547D+00,
     &.6570459382452019D-03,.7944457296146663D+00,.6579663080720945D-03,
     &.7968254443729997D+00,.6588636863112547D-03,.7991472777018847D+00,
     &.6597389584201900D-03,.8014133963335314D+00,.6605929718067311D-03,
     &.8036258773716017D+00,.6614265391084759D-03,.8057867159584249D+00,
     &.6622404414238483D-03,.8078978328788472D+00,.6630354315293254D-03/
       data (recf(i),i=3467,3521)/                .8099610821822974D+00,
     &.6638122371188470D-03,.8119782589091303D+00,.6645715641033481D-03,
     &.8139511070129370D+00,.6653141000107842D-03,.8158813275773382D+00,
     &.6660405175299570D-03,.8177705874338766D+00,.6667514782449204D-03,
     &.8196205282970527D+00,.6674476366107756D-03,.8214327765433879D+00,
     &.6681296442262653D-03,.8232089537737035D+00,.6687981544637654D-03,
     &.8249506883116352D+00,.6694538275230569D-03,.8266596278067956D+00,
     &.6700973359816298D-03,.8283374531279623D+00,.6707293709212047D-03,
     &.8299858937501699D+00,.6713506487175973D-03,.8316067448595304D+00,
     &.6719619185889133D-03,.8332018864208099D+00,.6725639710052013D-03,
     &.8347733044749573D+00,.6731576470708964D-03,.8363231149564427D+00,
     &.6737438489993667D-03,.8378535903427503D+00,.6743235518061943D-03,
     &.8393671894697208D+00,.6748978163539151D-03,.8408665908653167D+00,
     &.6754678038850311D-03,.8423547299689948D+00,.6760347921811532D-03,
     &.8438348406117731D+00,.6766001934827970D-03,.8453105011300979D+00,
     &.6771655742948708D-03,.8467856854705638D+00,.6777326771850360D-03,
     &.8482648196070496D+00,.6783034446530447D-03,.8497528435300649D+00,
     &.6788800451053436D-03,.8512552789714525D+00,.6794649009063182D-03,
     &.8527783028853924D+00,.6800607183903050D-03,.8543288265058851D+00/
       data (recf(i),i=3522,3576)/                .6806705196006733D-03,
     &.8559145795259218D+00,.6812976753667025D-03,.8575441985760543D+00,
     &.6819459391275817D-03,.8592273186991227D+00,.6826194806570309D-03,
     &.8609746659004745D+00,.6833229185231614D-03,.8627981480750792D+00,
     &.6840613497285638D-03,.8647109406514445D+00,.6848403745099397D-03,
     &.8667275621282101D+00,.6856661137341258D-03,.8688639333027147D+00,
     &.6865452157148281D-03,.8711374124074637D+00,.6874848486099405D-03,
     &.8735667966111717D+00,.6884926738773655D-03,.8761722784739423D+00,
     &.6895767956239815D-03,.8789753440908044D+00,.6907456801615706D-03,
     &.8819985980014631D+00,.6920080398012621D-03,.8852654987569412D+00,
     &.6933726750250028D-03,.8887999886800295D+00,.6948482698509175D-03,
     &.8926260022952612D+00,.6964431366613300D-03,.8967668406721245D+00,
     &.1687522084062742D-03,.2050350430521891D-03,.7143489225104815D-02,
     &.2367622652786059D-03,.3374351625166600D-01,.2645340906026387D-03,
     &.6835618366232366D-01,.2889390544558335D-03,.1056829987008023D+00,
     &.3104940979691790D-03,.1431437456375445D+00,.3296350786177901D-03,
     &.1794723727318731D+00,.3467237429868135D-03,.2140734485464243D+00,
     &.3620589217153205D-03,.2467046039766637D+00,.3758876341573480D-03,
     &.2773116025546536D+00,.3884147581578293D-03,.3059395516583021D+00/
       data (recf(i),i=3577,3631)/                .3998110227339366D-03,
     &.3326839408735750D+00,.4102194641152033D-03,.3576632785546656D+00,
     &.4197605903284232D-03,.3810037669469621D+00,.4285365009823915D-03,
     &.4028308242884812D+00,.4366341766023805D-03,.4232645539797606D+00,
     &.4441281128435874D-03,.4424175040394658D+00,.4510824389517647D-03,
     &.4603937534270471D+00,.4575526297479205D-03,.4772887576403662D+00,
     &.4635868963219365D-03,.4931896163747717D+00,.4692273217485011D-03,
     &.5081755622500646D+00,.4745107935135632D-03,.5223185511672184D+00,
     &.4794697730555439D-03,.5356838841319745D+00,.4841329341264508D-03,
     &.5483308203204026D+00,.4885256949620448D-03,.5603131593778051D+00,
     &.4926706640505523D-03,.5716797819868060D+00,.4965880152488166D-03,
     &.5824751443673352D+00,.5002958048415423D-03,.5927397262665235D+00,
     &.5038102406673260D-03,.6025104342019820D+00,.5071459114881099D-03,
     &.6118209628858234D+00,.5103159832377398D-03,.6207021182846412D+00,
     &.5133323675598979D-03,.6291821059168485D+00,.5162058670665758D-03,
     &.6372867879128392D+00,.5189463009622880D-03,.6450399121650791D+00,
     &.5215626140454308D-03,.6524633166362534D+00,.5240629715848113D-03,
     &.6595771116117261D+00,.5264548421517410D-03,.6663998424002746D+00,
     &.5287450701469129D-03,.6729486347171443D+00,.5309399394814304D-03/
       data (recf(i),i=3632,3686)/                .6792393247326799D+00,
     &.5330452296408964D-03,.6852865755411093D+00,.5350662651709710D-03,
     &.6911039815981985D+00,.5370079594647384D-03,.6967041624928251D+00,
     &.5388748536006167D-03,.7020988472546843D+00,.5406711508695714D-03,
     &.7072989502566063D+00,.5424007475382063D-03,.7123146396434907D+00,
     &.5440672603167729D-03,.7171553991088069D+00,.5456740509357201D-03,
     &.7218300837422168D+00,.5472242481790601D-03,.7263469705865458D+00,
     &.5487207676758462D-03,.7307138044675716D+00,.5501663297110800D-03,
     &.7349378395946277D+00,.5515634752832511D-03,.7390258773726449D+00,
     &.5529145806065207D-03,.7429843008159668D+00,.5542218702305272D-03,
     &.7468191059101640D+00,.5554874289292619D-03,.7505359302293567D+00,
     &.5567132124919160D-03,.7541400790825443D+00,.5579010575325747D-03,
     &.7576365494325465D+00,.5590526904217706D-03,.7610300518048510D+00,
     &.5601697354308826D-03,.7643250303805051D+00,.5612537221699207D-03,
     &.7675256814467769D+00,.5623060923901496D-03,.7706359703613220D+00,
     &.5633282062150818D-03,.7736596471697186D+00,.5643213478564601D-03,
     &.7766002610022336D+00,.5652867308658178D-03,.7794611733633313D+00,
     &.5662255029669265D-03,.7822455704165419D+00,.5671387505098367D-03,
     &.7849564743577130D+00,.5680275025831865D-03,.7875967539612229D+00/
       data (recf(i),i=3687,3741)/                .5688927348179374D-03,
     &.7901691343763252D+00,.5697353729126355D-03,.7926762062443136D+00,
     &.5705562959076292D-03,.7951204342015494D+00,.5713563392333756D-03,
     &.7975041648285145D+00,.5721362975559819D-03,.7998296341008703D+00,
     &.5728969274414421D-03,.8020989743949635D+00,.5736389498586027D-03,
     &.8043142210972835D+00,.5743630525397174D-03,.8064773188650033D+00,
     &.5750698922165000D-03,.8085901275829018D+00,.5757600967488519D-03,
     &.8106544280606480D+00,.5764342671629149D-03,.8126719275136132D+00,
     &.5770929796147660D-03,.8146442648700572D+00,.5777367872959363D-03,
     &.8165730159477090D+00,.5783662222969873D-03,.8184596985434300D+00,
     &.5789817974456180D-03,.8203057774808211D+00,.5795840081362118D-03,
     &.8221126696623233D+00,.5801733341683536D-03,.8238817491745800D+00,
     &.5807502416126773D-03,.8256143524986065D+00,.5813151847234263D-03,
     &.8273117838796568D+00,.5818686079183535D-03,.8289753209156332D+00,
     &.5824109478480398D-03,.8306062204274619D+00,.5829426355783928D-03,
     &.8322057246800982D+00,.5834640989120021D-03,.8337750680287443D+00,
     &.5839757648761716D-03,.8353154840714917D+00,.5844780624078450D-03,
     &.8368282133969516D+00,.5849714252682631D-03,.8383145120235179D+00,
     &.5854562952230521D-03,.8397756606357110D+00,.5859331255265216D-03/
       data (recf(i),i=3742,3796)/                .8412129747325485D+00,
     &.5864023847522156D-03,.8426278158130220D+00,.5868645610151844D-03,
     &.8440216037344282D+00,.5873201666349592D-03,.8453958303903652D+00,
     &.5877697432917324D-03,.8467520748664289D+00,.5882138677316644D-03,
     &.8480920202427393D+00,.5886531580803781D-03,.8494174722229419D+00,
     &.5890882808263732D-03,.8507303797787185D+00,.5895199585379871D-03,
     &.8520328580063178D+00,.5899489783783064D-03,.8533272133961881D+00,
     &.5903762014816018D-03,.8546159717171726D+00,.5908025732518324D-03,
     &.8559019087112308D+00,.5912291346377758D-03,.8571880837811911D+00,
     &.5916570344294526D-03,.8584778768299521D+00,.5920875426055483D-03,
     &.8597750283715609D+00,.5925220647400795D-03,.8610836829786467D+00,
     &.5929621574468788D-03,.8624084360518852D+00,.5934095448005555D-03,
     &.8637543837896255D+00,.5938661356200435D-03,.8651271760926207D+00,
     &.5943340414329835D-03,.8665330719519609D+00,.5948155948529893D-03,
     &.8679789966288171D+00,.5953133679941353D-03,.8694725996326263D+00,
     &.5958301904145430D-03,.8710223121296252D+00,.5963691659207649D-03,
     &.8726374019561603D+00,.5969336873744813D-03,.8743280238622816D+00,
     &.5975274484218217D-03,.8761052619650037D+00,.5981544508145503D-03,
     &.8779811606467382D+00,.5988190057157910D-03,.8799687393004150D+00/
       data (recf(i),i=3797,3851)/                .5995257270899495D-03,
     &.8820819854186045D+00,.6002795149823573D-03,.8843358195863941D+00,
     &.6010855262223838D-03,.8867460250263935D+00,.6019491298677465D-03,
     &.8893291335470804D+00,.6028758445922773D-03,.8921022591848045D+00,
     &.6038712552612013D-03,.8950828706648276D+00,.6049409062046725D-03,
     &.8982884942350033D+00,.6060901692668508D-03,.9017363396754307D+00,
     &.1481251565099711D-03,.1799742621307956D-03,.7143974934319647D-02,
     &.2078249765617697D-03,.3374626735466673D-01,.2322040770469374D-03,
     &.6836267972058510D-01,.2536281024389343D-03,.1056944648470500D+00,
     &.2725507603509361D-03,.1431611968300346D+00,.2893545952483257D-03,
     &.1794966519846322D+00,.3043571067937865D-03,.2141052583069300D+00,
     &.3178205747305427D-03,.2467445313185863D+00,.3299618049127232D-03,
     &.2773601392459256D+00,.3409606161517330D-03,.3059971101361032D+00,
     &.3509668551995601D-03,.3327508670011682D+00,.3601060630945735D-03,
     &.3577398618615326D+00,.3684840080409944D-03,.3810902489006329D+00,
     &.3761903012944965D-03,.4029274050222500D+00,.3833012841788339D-03,
     &.4233713978033811D+00,.3898823401138323D-03,.4425347439853034D+00,
     &.3959897539743794D-03,.4605214950272289D+00,.4016722146902933D-03,
     &.4774270820870889D+00,.4069720358522181D-03,.4933385831828130D+00/
       data (recf(i),i=3852,3906)/                .4119261525257769D-03,
     &.5083352115151812D+00,.4165669396400193D-03,.5224889054945915D+00,
     &.4209228874124991D-03,.5358649502937634D+00,.4250191616382618D-03,
     &.5485225906897057D+00,.4288780707755034D-03,.5605156131771370D+00,
     &.4325194571969843D-03,.5718928863819531D+00,.4359610264298928D-03,
     &.5826988554317172D+00,.4392186254392934D-03,.5929739898366225D+00,
     &.4423064788406665D-03,.6027551866407443D+00,.4452373902181403D-03,
     &.6120761317679944D+00,.4480229143725210D-03,.6209676230156103D+00,
     &.4506735052476993D-03,.6294578582946426D+00,.4531986434246486D-03,
     &.6375726926413187D+00,.4556069463824008D-03,.6453358673250855D+00,
     &.4579062641690959D-03,.6527692141203824D+00,.4601037626756183D-03,
     &.6598928375275065D+00,.4622059963377751D-03,.6667252774457851D+00,
     &.4642189717935211D-03,.6732836545324753D+00,.4661482037761131D-03,
     &.6795838002301146D+00,.4679987643218002D-03,.6856403732164480D+00,
     &.4697753262034514D-03,.6914669638252581D+00,.4714822013627870D-03,
     &.6970761878028088D+00,.4731233749983658D-03,.7024797706018147D+00,
     &.4747025358699538D-03,.7076886232711463D+00,.4762231032989864D-03,
     &.7127129108730295D+00,.4776882512767774D-03,.7175621142484538D+00,
     &.4791009300347118D-03,.7222450858541227D+00,.4804638853820701D-03/
       data (recf(i),i=3907,3961)/                .7267701003089459D+00,
     &.4817796760758937D-03,.7311449002133155D+00,.4830506894522046D-03,
     &.7353767377389203D+00,.4842791555179398D-03,.7394724124294618D+00,
     &.4854671596773276D-03,.7434383056023196D+00,.4866166542444491D-03,
     &.7472804116970708D+00,.4877294688748149D-03,.7510043668780048D+00,
     &.4888073200324911D-03,.7546154751637144D+00,.4898518195952196D-03,
     &.7581187323268813D+00,.4908644826877864D-03,.7615188477809879D+00,
     &.4918467348233061D-03,.7648202646474365D+00,.4927999184228891D-03,
     &.7680271781760355D+00,.4937252987761405D-03,.7711435526736958D+00,
     &.4946240694979417D-03,.7741731370801548D+00,.4954973575308488D-03,
     &.7771194793153748D+00,.4963462277370809D-03,.7799859395107068D+00,
     &.4971716871193750D-03,.7827757022247849D+00,.4979746887058557D-03,
     &.7854917877352522D+00,.4987561351304407D-03,.7881370624886681D+00,
     &.4995168819371097D-03,.7907142487831792D+00,.5002577406335545D-03,
     &.7932259337516524D+00,.5009794815172498D-03,.7956745777068556D+00,
     &.5016828362948059D-03,.7980625219048639D+00,.5023685005135441D-03,
     &.8003919957780823D+00,.5030371358225455D-03,.8026651236850512D+00,
     &.5036893720789472D-03,.8048839312204863D+00,.5043258093139596D-03,
     &.8070503511257529D+00,.5049470195719527D-03,.8091662288371395D+00/
       data (recf(i),i=3962,4016)/                .5055535486349787D-03,
     &.8112333277068590D+00,.5061459176442604D-03,.8132533339296184D+00,
     &.5067246246294598D-03,.8152278612058557D+00,.5072901459559465D-03,
     &.8171584551713140D+00,.5078429376997978D-03,.8190465976214947D+00,
     &.5083834369598832D-03,.8208937105586977D+00,.5089120631160979D-03,
     &.8227011600888023D+00,.5094292190426258D-03,.8244702601946616D+00,
     &.5099352922850141D-03,.8262022764129838D+00,.5104306562098391D-03,
     &.8278984294418413D+00,.5109156711358289D-03,.8295598987064962D+00,
     &.5113906854554846D-03,.8311878259120602D+00,.5118560367565149D-03,
     &.8327833186126263D+00,.5123120529527563D-03,.8343474538279233D+00,
     &.5127590534347159D-03,.8358812817402723D+00,.5131973502504269D-03,
     &.8373858295066677D+00,.5136272493279681D-03,.8388621052231878D+00,
     &.5140490517517615D-03,.8403111020816669D+00,.5144630551056335D-03,
     &.8417338027616508D+00,.5148695548966082D-03,.8431311841041218D+00,
     &.5152688460744951D-03,.8445042221173288D+00,.5156612246635426D-03,
     &.8458538973693004D+00,.5160469895237514D-03,.8471812008262565D+00,
     &.5164264442608728D-03,.8484871402011642D+00,.5167998993056541D-03,
     &.8497727468820885D+00,.5171676741845232D-03,.8510390835157447D+00,
     &.5175301000056107D-03,.8522872523277160D+00,.5178875221857676D-03/
       data (recf(i),i=4017,4071)/                .8535184042670900D+00,
     &.5182403034460151D-03,.8547337490696901D+00,.5185888271046153D-03,
     &.8559345663404898D+00,.5189335006986103D-03,.8571222177620127D+00,
     &.5192747599661662D-03,.8582981605412726D+00,.5196130732232652D-03,
     &.8594639622127785D+00,.5199489461690734D-03,.8606213169188581D+00,
     &.5202829271545004D-03,.8617720632905069D+00,.5206156129478311D-03,
     &.8629182040514101D+00,.5209476550295670D-03,.8640619274638366D+00,
     &.5212797664454193D-03,.8652056307266341D+00,.5216127292413131D-03,
     &.8663519454211874D+00,.5219474024967696D-03,.8675037650792340D+00,
     &.5222847309624981D-03,.8686642749148173D+00,.5226257542936992D-03,
     &.8698369837189167D+00,.5229716168515654D-03,.8710257578565097D+00,
     &.5233235780207398D-03,.8722348572285707D+00,.5236830229588920D-03,
     &.8734689729618574D+00,.5240514736547882D-03,.8747332664628638D+00,
     &.5244306001218837D-03,.8760334093141518D+00,.5248222314940908D-03,
     &.8773756232962860D+00,.5252283667175941D-03,.8787667195815746D+00,
     &.5256511844461572D-03,.8802141358619505D+00,.5260930516464654D-03,
     &.8817259698386895D+00,.5265565303044835D-03,.8833110071141399D+00,
     &.5270443814944088D-03,.8849787410860012D+00,.5275595659308863D-03,
     &.8867393819581020D+00,.5281052399771031D-03,.8886038514594998D+00/
       data (recf(i),i=4072,4126)/                .5286847459332971D-03,
     &.8905837593259870D+00,.5293015952926587D-03,.8926913570756628D+00,
     &.5299594435392922D-03,.8949394641477838D+00,.5306620549953031D-03,
     &.8973413611324083D+00,.5314132562256762D-03,.8999106446760794D+00,
     &.5322168766096216D-03,.9026610388027025D+00,.5330766749183085D-03,
     &.9056061579514983D+00,.5339962511354281D-03,.9087592181279729D+00,
     &.1291672658807919D-03,.1569411415132314D-03,.7144429374444041D-02,
     &.1812286696375032D-03,.3374884163870904D-01,.2024891487895646D-03,
     &.6836875906346668D-01,.2211729609091939D-03,.1057051970576302D+00,
     &.2376757078934563D-03,.1431775337300160D+00,.2523309440420044D-03,
     &.1795193851821314D+00,.2654155096114621D-03,.2141350485529856D+00,
     &.2771580968905830D-03,.2467819320733997D+00,.2877477478365611D-03,
     &.2774056154477370D+00,.2973412539386665D-03,.3060510531385813D+00,
     &.3060692727883761D-03,.3328136064982617D+00,.3140412687616450D-03,
     &.3578116754563008D+00,.3213494654089220D-03,.3811713698313008D+00,
     &.3280719982928522D-03,.4030180284424702D+00,.3342754323013638D-03,
     &.4234716859228214D+00,.3400167776073826D-03,.4426448303068433D+00,
     &.3453451109288780D-03,.4606414878637722D+00,.3503028857160415D-03,
     &.4775570675004032D+00,.3549269964559184D-03,.4934786274558236D+00/
       data (recf(i),i=4127,4181)/                .3592496478431822D-03,
     &.5084853632454121D+00,.3632990683734921D-03,.5226491973781846D+00,
     &.3671000992804959D-03,.5360354006553708D+00,.3706746830802694D-03,
     &.5487032048025572D+00,.3740422708473267D-03,.5607063844089743D+00,
     &.3772201633670793D-03,.5720937971957492D+00,.3802237982173855D-03,
     &.5829098782647340D+00,.3830669924186714D-03,.5931950878770883D+00,
     &.3857621484003172D-03,.6029863145180096D+00,.3883204295409269D-03,
     &.6123172361692740D+00,.3907519103608007D-03,.6212186432401899D+00,
     &.3930657055071256D-03,.6297187267546153D+00,.3952700809230814D-03,
     &.6378433353164356D+00,.3973725499905643D-03,.6456162041780895D+00,
     &.3993799569511753D-03,.6530591594781886D+00,.4012985495172340D-03,
     &.6601923004327613D+00,.4031340422649590D-03,.6670341619826434D+00,
     &.4048916721408514D-03,.6736018601298540D+00,.4065762471981460D-03,
     &.6799112219451941D+00,.4081921895038212D-03,.6859769020007800D+00,
     &.4097435730108617D-03,.6918124867754874D+00,.4112341570695012D-03,
     &.6974305883977158D+00,.4126674161504436D-03,.7028429289271262D+00,
     &.4140465662688992D-03,.7080604162333366D+00,.4153745885277149D-03,
     &.7130932124031323D+00,.4166542501385347D-03,.7179507954967295D+00,
     &.4178881232298605D-03,.7226420153762622D+00,.4190786017085115D-03/
       data (recf(i),i=4182,4236)/                .7271751442443414D+00,
     &.4202279164050269D-03,.7315579224557873D+00,.4213381487029441D-03,
     &.7357976001001488D+00,.4224112428257702D-03,.7399009747952365D+00,
     &.4234490169331078D-03,.7438744260815747D+00,.4244531731582231D-03,
     &.7477239467635279D+00,.4254253067028491D-03,.7514551715040853D+00,
     &.4263669140908001D-03,.7550734029462083D+00,.4272794006696848D-03,
     &.7585836356036648D+00,.4281640874393643D-03,.7619905777378661D+00,
     &.4290222172765655D-03,.7652986714139369D+00,.4298549606170237D-03,
     &.7685121109086968D+00,.4306634206495281D-03,.7716348596250679D+00,
     &.4314486380701255D-03,.7746706656513523D+00,.4322115954393903D-03,
     &.7776230760895890D+00,.4329532211809737D-03,.7804954502645794D+00,
     &.4336743932555310D-03,.7832909719139642D+00,.4343759425404994D-03,
     &.7860126604497781D+00,.4350586559430094D-03,.7886633813730524D+00,
     &.4357232792703970D-03,.7912458559151491D+00,.4363705198802944D-03,
     &.7937626699724827D+00,.4370010491300795D-03,.7962162823950203D+00,
     &.4376155046435118D-03,.7986090326833559D+00,.4382144924106579D-03,
     &.8009431481441662D+00,.4387985887356742D-03,.8032207505493939D+00,
     &.4393683420456565D-03,.8054438623405243D+00,.4399242745725549D-03,
     &.8076144124157677D+00,.4404668839190812D-03,.8097342415347859D+00/
       data (recf(i),i=4237,4291)/                .4409966445185832D-03,
     &.8118051073727768D+00,.4415140089980146D-03,.8138286892532131D+00,
     &.4420194094523807D-03,.8158065925862987D+00,.4425132586383796D-03,
     &.8177403530382282D+00,.4429959510943727D-03,.8196314404545940D+00,
     &.4434678641933102D-03,.8214812625597571D+00,.4439293591347870D-03,
     &.8232911684526717D+00,.4443807818820194D-03,.8250624519185151D+00,
     &.4448224640492010D-03,.8267963545745015D+00,.4452547237444166D-03,
     &.8284940688674631D+00,.4456778663730607D-03,.8301567409401316D+00,
     &.4460921854065215D-03,.8317854733825585D+00,.4464979631207512D-03,
     &.8333813278847675D+00,.4468954713092401D-03,.8349453278065194D+00,
     &.4472849719748584D-03,.8364784606800110D+00,.4476667180050050D-03,
     &.8379816806613994D+00,.4480409538345297D-03,.8394559109472592D+00,
     &.4484079161009533D-03,.8409020461724398D+00,.4487678342966111D-03,
     &.8423209548062871D+00,.4491209314224874D-03,.8437134815648469D+00,
     &.4494674246486925D-03,.8450804498574688D+00,.4498075259867541D-03,
     &.8464226642871850D+00,.4501414429791688D-03,.8477409132253626D+00,
     &.4504693794119635D-03,.8490359714824147D+00,.4507915360563808D-03,
     &.8503086030978187D+00,.4511081114462013D-03,.8515595642743328D+00,
     &.4514193026976681D-03,.8527896064831288D+00,.4517253063794774D-03/
       data (recf(i),i=4292,4346)/                .8539994797685749D+00,
     &.4520263194408489D-03,.8551899362836138D+00,.4523225402062811D-03,
     &.8563617340890786D+00,.4526141694462433D-03,.8575156412528867D+00,
     &.4529014115337412D-03,.8586524402878235D+00,.4531844756974208D-03,
     &.8597729329695807D+00,.4534635773826384D-03,.8608779455798124D+00,
     &.4537389397327130D-03,.8619683346221994D+00,.4540107952033793D-03,
     &.8630449930628203D+00,.4542793873242631D-03,.8641088571494663D+00,
     &.4545449726219747D-03,.8651609138678200D+00,.4548078227201481D-03,
     &.8662022090955665D+00,.4550682266323894D-03,.8672338565183648D+00,
     &.4553264932646083D-03,.8682570473740338D+00,.4555829541435220D-03,
     &.8692730610930815D+00,.4558379663881761D-03,.8702832769045586D+00,
     &.4560919159410250D-03,.8712891864758329D+00,.4563452210743556D-03,
     &.8722924076528336D+00,.4565983361864682D-03,.8732946993631147D+00,
     &.4568517558999058D-03,.8742979777371169D+00,.4571060194709263D-03,
     &.8753043334925269D+00,.4573617155151244D-03,.8763160506117581D+00,
     &.4576194870483323D-03,.8773356263222484D+00,.4578800368343567D-03,
     &.8783657923622515D+00,.4581441330213403D-03,.8794095374796469D+00,
     &.4584126150361641D-03,.8804701310663527D+00,.4586863996908342D-03,
     &.8815511477743261D+00,.4589664874357118D-03,.8826564928888149D+00/
       data (recf(i),i=4347,4382)/                .4592539686711794D-03,
     &.8837904281482389D+00,.4595500300013319D-03,.8849575975954920D+00,
     &.4598559602799827D-03,.8861630529202465D+00,.4601731562602071D-03,
     &.8874122776038270D+00,.4605031276134677D-03,.8887112090056589D+00,
     &.4608475010329713D-03,.8900662573321374D+00,.4612080230785231D-03,
     &.8914843202051330D+00,.4615865613575363D-03,.8929727913000981D+00,
     &.4619851035704269D-03,.8945395612571059D+00,.4624057538807515D-03,
     &.8961930087896581D+00,.4628507260045860D-03,.8979419796374420D+00,
     &.4633223323547168D-03,.8997957507473553D+00,.4638229685297288D-03,
     &.9017639768452591D+00,.4643550924143431D-03,.9038566164094044D+00,
     &.4649211971655587D-03,.9060838340132801D+00,.4655237774111018D-03,
     &.9084558761162183D+00,.4661652880954744D-03,.9109829176962647D+00,
     &.4668480955879464D-03,.9136748776971255D+00/
c *** *** *** *** *** *** *** *** *** *** *** *** *** *** ***
c             initializations
c *** *** *** *** *** *** *** *** *** *** *** *** *** *** ***
      arret=.false.
c -------- Prepare the entry-points for the arrays in work.--------
      n1=1
      n2=n1+neqn
      n3=n2+neqn
      n4=n3+neqn
      n5=n4+neqn
      n6=n5+neqn
      n7=n6+neqn
c -------- Test the initial step size and tolerances.--------
      if (h.gt.abs(tend-t)) then
        write(6,*)'initial step is longer than the integration interval'
        arret=.true.
        idid=-1
        return
      end if
      if (h.lt.10.d0*uround) then
        write(6,*)'initial step-size is too small'
        idid=-1
        arret=.true.
        return
      end if
      if (iwork(4).eq.0) then
        ntol=0
        if (atol(1).le.0.d0.or.rtol(1).le.10.d0*uround) then
          write(6,*) 'tolerances are too small'
          arret=.true.
          idid=-1
          return
        end if
      else
        ntol=1
        do i=1,neqn
        if (atol(i).le.0.d0.or.rtol(i).le.10.d0*uround) then
          write(6,*) 'tolerances are too small'
          arret=.true.
          idid=-1
          return
        end if
        end do
      end if
c -------- Call to the core integrator.-----------
      call rockfcore(neqn,t,tend,h,y,f,work,work(n1),work(n2),
     & work(n3),work(n4),work(n5),work(n6),work(n7),ms,atol,rtol,
     & ntol,recf,fpa,fpb,fpbe,iwork,arret,uround,idid)
      return
      end
c ----------------------------------------------
c     End of subroutine ROCK4.
c ----------------------------------------------
c
       subroutine rockfcore(neqn,t,tend,h,y,f,work,yn,fn,fnt,yjm1,
     & yjm2,yjm3,yjm4,ms,atol,rtol,ntol,recf,fpa,fpb,fpbe,iwork,
     & arret,uround,idid)
c ----------------------------------------------
c    Core integrator for ROCK4.
c ----------------------------------------------
c             Declarations
c ----------------------------------------------
       double precision y(*),yn(*),fn(*),fnt(neqn),work(*),
     & yjm1(neqn),yjm2(neqn),yjm3(neqn),yjm4(neqn),recf(4382),
     & atol(*),rtol(*),fpa(50,6),fpb(50,4),fpbe(50,5),err,errp,
     & eigmax,t,tend,h,hnew,hp,facmax,fac,facp,rho,uround,told
       integer ms(50),mp(2),iwork(12),neqn,mdeg,i,ntol,
     & nrho,mdego,nrej,idid
       logical last,reject,arret
       external f
c *** *** *** *** *** *** *** *** *** *** *** *** *** *** ***
c             initializations
c *** *** *** *** *** *** *** *** *** *** *** *** *** *** ***
      iwork(5)=0
      iwork(6)=0
      iwork(7)=0
      iwork(8)=0
      iwork(9)=0
      iwork(10)=0
      iwork(11)=0
      iwork(12)=0
      facmax=5.d0
      told=0.d0
      nrej=0
      hp=h
      mdego=0
      last=.false.
      reject=.false.
      err=0.d0
      nrho=0
      idid=1
c -------- initialization of the integration step.--------
10    if (iwork(6).ne.0) then
        do i=1,neqn
          yn(i)=y(i)
          fn(i)=fnt(i)
        end do
               else
        do i=1,neqn
          yn(i)=y(i)
        end do
      call f(neqn,t,yn,fn)
      end if
      iwork(5)=iwork(5)+1
      errp=err
c -------- Step-size is adjusted.--------
20    if(1.01d0*h.ge.abs(tend-t)) then
        h=dabs(tend-t)
        last=.true.
      end if
      if (h.lt.10.d0*uround) then
        write(6,*)' tolerances are too small'
        idid=-2
        arret=.true.
        return
      end if
c -------- Spectral radius.--------
      if (nrho.eq.0) then
        if (iwork(6).eq.0.or.iwork(2).eq.0) then
c ------- Computed externally by rho.--------
          if (iwork(1).eq.1) then
            eigmax=rho(neqn,t,yn)
            if (idnint(eigmax).gt.iwork(11))
     &         iwork(11)=idnint(eigmax)
            if (iwork(6).eq.0) iwork(12)=iwork(11)
            if (idnint(eigmax).lt.iwork(12))
     &         iwork(12)=idnint(eigmax)
c ------- Computed internally by rockfrho.--------
                             else
            call rockfrho(neqn,t,y,f,yn,fn,work,yjm1,
     &                   yjm2,eigmax,uround,idid,iwork)
            if (idnint(eigmax).gt.iwork(11))
     &         iwork(11)=idnint(eigmax)
            if (iwork(6).eq.0) iwork(12)=iwork(11)
            if (idnint(eigmax).lt.iwork(12))
     &         iwork(12)=idnint(eigmax)
          end if
        end if
      end if
c -------- The number of stages.--------
       mdeg=sqrt((3.d0+h*eigmax)/0.353d0)+1
      if (mdeg.gt.152) then
        h=0.8d0*(152.d0**2*0.353d0-3.d0)/eigmax
        mdeg=152
        last=.false.
      end if
      mdeg=max(mdeg,5)-4
      if (mdeg.ne.mdego) then
        call mdegre(mdeg,mp,ms)
      end if
      if (mdeg+4.gt.iwork(10)) iwork(10)=mdeg+4
c -------- Computation of an integration step.--------
      call rfstep(neqn,t,h,y,f,yn,fn,fnt,yjm1,yjm2,yjm3,yjm4,
     & mdeg,mp,err,atol,rtol,ntol,recf,fpa,fpb,fpbe)
      mdego=mdeg
      iwork(6)=iwork(6)+1
      iwork(5)=iwork(5)+mdeg+4
c -------- Error control procedure.--------
      fac=(1.d0/err)**(0.25d0)
      if (errp.ne.0.d0.and..not.reject) then
        facp=errp**(0.25d0)*fac**2*(h/hp)
        fac=dmin1(fac,facp)
      end if
      if (reject) then
        facmax=1.d0
      end if
      fac=dmin1(facmax,dmax1(0.1d0,0.8d0*fac))
      hnew=h*fac
c -------- Accepted step.--------
      if (err.lt.1.d0) then
        iwork(7)=iwork(7)+1
        facmax=2.d0
        t=t+h
        if (reject) then
          hnew=dmin1(hnew,h)
          if (tend.lt.t)  hnew=dmax1(hnew,h)
          reject= .false.
          nrej=0
        end if
        hp=h
        h=hnew
        nrho=nrho+1
        nrho=mod(nrho+1,10)
        if (last) then
          return
        elseif (iwork(3).eq.1) then
          if (idid.eq.1) idid=2
          if(1.1d0*h.ge.abs(tend-t)) h=dabs(tend-t)
          return
        else
          goto 10
        end if
                   else
c -------- Rejected step.--------
        iwork(8)=iwork(8)+1
        reject= .true.
        last=.false.
        h= 0.8d0*hnew
        if (iwork(6).eq.0) h=0.1d0*h
        if (told.eq.t) then
          nrej=nrej+1
          if (nrej.eq.10) h=1.0d-5
        end if
        told=t
c -------The spectral radius is recomputed.--------
c        after a step failure
        if (nrho.ne.0) then
          nrho=0
                       else
          nrho=1
        end if
        goto 20
      end if
      return
      end
c ----------------------------------------------
c     End of subroutine rockcore.
c ----------------------------------------------
c
      subroutine rfstep(neqn,t,h,y,f,yn,fn,fnt,yjm1,yjm2,yjm3,
     & yjm4,mdeg,mp,err,atol,rtol,ntol,recf,fpa,fpb,fpbe)
c ----------------------------------------------
c  Solut. at t+h by an explicit (mdeg+4)-stages formula.
c-----------------------------------------------
c ----------------------------------------------
c             Declarations
c ----------------------------------------------
       double precision y(neqn),yn(neqn),fn(neqn),fnt(neqn),
     & yjm2(neqn),yjm3(neqn),yjm4(neqn),recf(4382),fpa(50,6),
     & yjm1(neqn),fpb(50,4),fpbe(50,5),atol(*),rtol(*),err,t,
     & h,ci1,ci2,ci3,temp1,temp2,temp3,temp4,temp5,ato,rto
       integer mp(2),neqn,mdeg,mr,mz,i,j,ntol
       external f
c *** *** *** *** *** *** *** *** *** *** *** *** *** *** ***
c             initializations
c *** *** *** *** *** *** *** *** *** *** *** *** *** *** ***
      err=0.d0
      mz=mp(1)
      mr=mp(2)
c -------- First stage.--------
      temp1=h*recf(mr)
      ci1=t+temp1
      ci2=t+temp1
      ci3=t
      do i=1,neqn
        yjm2(i)=yn(i)
        yjm1(i)=yn(i)+temp1*fn(i)
        if (mdeg.lt.2) then
          y(i)=yjm1(i)
        end if
      end do
c --------- Stage for j=2..mdeg.--------
      do i=2,mdeg
        temp1=h*recf(mr+2*(i-2)+1)
        temp3=-recf(mr+2*(i-2)+2)
        temp2=1.d0-temp3
        call f(neqn,ci1,yjm1,y)
        ci1=temp1+temp2*ci2+temp3*ci3
        do j=1,neqn
          y(j)=temp1*y(j)+temp2*yjm1(j)+temp3*yjm2(j)
c -------- Shift the value "y" for the next stage.--------
          if (i.lt.mdeg) then
            yjm2(j)=yjm1(j)
            yjm1(j)=y(j)
          end if
        end do
        ci3=ci2
        ci2=ci1
      end do
c -------- The finishing procedure (4-stage method).
c -------- Stage 1.--------
      temp1=h*fpa(mz,1)
      call f(neqn,ci1,y,yjm1)
      do j=1,neqn
        yjm3(j)=y(j)+temp1*yjm1(j)
      end do
c -------- Stage 2.--------
      ci2=ci1+temp1
      temp1=h*fpa(mz,2)
      temp2=h*fpa(mz,3)
      call f(neqn,ci2,yjm3,yjm2)
      do j=1,neqn
        yjm4(j)=y(j)+temp1*yjm1(j)+temp2*yjm2(j)
      end do
c -------- Stage 3.--------
      ci2=ci1+temp1+temp2
      temp1=h*fpa(mz,4)
      temp2=h*fpa(mz,5)
      temp3=h*fpa(mz,6)
      call f(neqn,ci2,yjm4,yjm3)
      do j=1,neqn
        fnt(j)=y(j)+temp1*yjm1(j)+temp2*yjm2(j)+temp3*yjm3(j)
      end do
c -------- Stage 4.--------
      ci2=ci1+temp1+temp2+temp3
      temp1=h*fpb(mz,1)
      temp2=h*fpb(mz,2)
      temp3=h*fpb(mz,3)
      temp4=h*fpb(mz,4)
      call f(neqn,ci2,fnt,yjm4)
      do j=1,neqn
        y(j)=y(j)+temp1*yjm1(j)+temp2*yjm2(j)+temp3*yjm3(j)
     &  +temp4*yjm4(j)
      end do
c -------- Error evaluation (embedded method of order 3).--------
      temp1=h*fpbe(mz,1)-temp1
      temp2=h*fpbe(mz,2)-temp2
      temp3=h*fpbe(mz,3)-temp3
      temp4=h*fpbe(mz,4)-temp4
      temp5=h*fpbe(mz,5)
      call f(neqn,t+h,y,fnt)
c -------- Atol and rtol are scalar.--------
      if (ntol.eq.0) then
        ato=atol(1)
        rto=rtol(1)
        do j=1,neqn
          ci1=dabs(y(j))*rto
          err=err+((temp1*yjm1(j)+temp2*yjm2(j)+temp3*yjm3(j)
     &   +temp4*yjm4(j)+temp5*fnt(j))/(ato+ci1))**2
        end do
c -------- Atol and rtol are array.--------
                     else
        do j=1,neqn
          ci1=dabs(y(j))*rtol(j)
          err=err+((temp1*yjm1(j)+temp2*yjm2(j)+temp3*yjm3(j)
     &   +temp4*yjm4(j)+temp5*fnt(j))/(atol(j)+ci1))**2
        end do
      end if
      err=sqrt(err/neqn)
      return
      end
c ----------------------------------------------
c     End of subroutine rfstep.
c ----------------------------------------------
c
      subroutine mdegre(mdeg,mp,ms)
c-------------------------------------------------------------
c       Find the optimal degree.
c       MP(1): pointer which select the degree in ms(i)\1,2,..
c             such that mdeg<=ms(i).
c       MP(2): pointer which gives the corresponding position
c       of a_1 in the data recf for the selected degree.
c-------------------------------------------------------------
c ----------------------------------------------
c             Declarations
c ----------------------------------------------
      integer ms(50),mp(2),mdeg,i
c *** *** *** *** *** *** *** *** *** *** *** *** *** *** ***
c             initializations
c *** *** *** *** *** *** *** *** *** *** *** *** *** *** ***
      mp(2)=1
c -------- Find the degree.--------
      do i=1,50
        if ((ms(i)/mdeg).ge.1) then
          mdeg=ms(i)
          mp(1)=i
          return
        end if
      mp(2)=mp(2)+ms(i)*2-1
      end do
      return
      end
c ----------------------------------------------
c     End of subroutine mdegr.
c ----------------------------------------------
c
      subroutine rockfrho(neqn,t,y,f,yn,fn,work,z,fz,eigmax,
     &                   uround,idid,iwork)
c------------------------------------------------------------
c     Rockfrho compute eigmax, a close upper bound of the
c     spectral radius of the Jacobian matrix using a
c     power method (J.N. Franklin (matrix theory)).
c     The algorithm used is a small change (initial vector
c     and stopping criteria) of that of
c     Sommeijer-Shampine-Verwer, implemented in RKC.
c-------------------------------------------------------------
c             Declarations
c-------------------------------------------------------------
       double precision y(neqn),yn(neqn),fn(neqn),z(neqn),
     & work(*),fz(neqn),t,eigmax,eigmaxo,sqrtu,uround,znor,
     & ynor,quot,dzyn,dfzfn,safe
       integer iwork(12),neqn,n8,i,iter,maxiter,nind,ntest,
     & ind,idid
       parameter (maxiter=50)
       parameter (safe=1.2d0)
       external f
c *** *** *** *** *** *** *** *** *** *** *** *** *** *** ***
c             initializations
c *** *** *** *** *** *** *** *** *** *** *** *** *** *** ***
      sqrtu=sqrt(uround)
      ynor=0.d0
      znor=0.d0
      n8=7*neqn
c ------ The initial vectors for the power method are yn --------
c       and yn+c*f(v_n), where vn=f(yn) a perturbation of yn
c       (if iwork(6)=0) or a perturbation of the last computed
c       eigenvector (if iwork(6).neq.0).
c
      if (iwork(6).eq.0) then
        do i=1,neqn
          fz(i)=fn(i)
        end do
        call f(neqn,t,fz,z)
        iwork(9)=iwork(9)+1
                         else
        do i=1,neqn
          z(i)=work(n8+i)
        end do
      end if
c ------ Perturbation.--------
      do i=1,neqn
        ynor=ynor+yn(i)**2
        znor=znor+z(i)**2
      end do
      ynor=sqrt(ynor)
      znor=sqrt(znor)
c ------ Normalization of the vector z so that --------
c        the difference z-yn lie in a circle
c        around yn (i.e has a constant modulus).
c
      if (ynor.ne.0.d0.and.znor.ne.0.d0) then
        dzyn=ynor*sqrtu
        quot=dzyn/znor
        do i=1,neqn
          z(i)=yn(i)+z(i)*quot
        end do
      elseif(ynor.ne.0.d0) then
        dzyn=ynor*sqrtu
        do i=1,neqn
          z(i)=yn(i)+yn(i)*sqrtu
        end do
      elseif(znor.ne.0.d0) then
          dzyn=uround
          quot=dzyn/znor
          do i=1,neqn
            z(i)=z(i)*quot
          end do
      else
        dzyn=uround
        do i=1,neqn
          z(i)=dzyn
        end do
      end if
c ------ Start the power method.--------
      eigmax=0.d0
      do iter=1,maxiter
        call f(neqn,t,z,fz)
        iwork(9)=iwork(9)+1
        dfzfn=0.d0
        do i=1,neqn
          dfzfn=dfzfn+(fz(i)-fn(i))**2
        end do
        dfzfn=dsqrt(dfzfn)
        eigmaxo=eigmax
        eigmax=dfzfn/dzyn
        eigmax=safe*eigmax
c ------ The stopping criteria is based on a
c        relative error between two succesive
c        estimation ``eigmax'' of the spectral
c        radius.
c
        if (iter.ge.2.and.dabs(eigmax-eigmaxo)
     &    .le.(eigmax*0.05d0)) then
c ----- The last eigenvector is stored.--------
          do i=1,neqn
            work(n8+i)=z(i)-yn(i)
          end do
          return
        end if
c ----- The next z is defined by --------
c       z_new=yn+coef*(fz-fn) where
c       coef is choosen so that
c       norm(z_new-yn)=norm(z_old-yn)
c
        if (dfzfn.ne.0.d0) then
          quot=dzyn/dfzfn
          do i=1,neqn
            z(i)=yn(i)+(fz(i)-fn(i))*quot
          end do
        else
c ----- The new z is defined by an arbitrary --------
c       perturbation of the current approximation
c       of the eigenvector.
c
          nind=neqn
          ntest=0
          ind=1+mod(iter,nind)
          if (z(ind).ne.yn(ind).or.ntest.eq.10) then
            z(ind)=yn(ind)-(z(ind)-yn(ind))
          else
            nind=neqn+ind
            ntest=ntest+1
          end if
        end if
      end do
      write(6,*) 'convergence failure in the
     & spectral radius computation'
      idid=-3
      return
      end
